Package: TENxIO
Type: Package
Title: Import methods for 10X Genomics files
Version: 1.14.0
Authors@R: c(
    person("Marcel", "Ramos", , "marcel.ramos@sph.cuny.edu", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-3242-0582")),
    person("NCI", role = "fnd",
           comment = c(GrantNo. = "U24CA289073"))
  )
Depends: R (>= 4.5.0), SingleCellExperiment, SummarizedExperiment
Imports: BiocBaseUtils, BiocGenerics, BiocIO, Seqinfo, GenomicRanges,
        HDF5Array, Matrix, MatrixGenerics, methods, readr, rhdf5,
        R.utils, S4Vectors, utils
Suggests: BiocStyle, DropletTestFiles, ExperimentHub, knitr,
        RaggedExperiment (>= 1.33.3), rmarkdown, Rsamtools, tinytest
Description: Provides a structured S4 approach to importing data files
        from the 10X pipelines. It mainly supports Single Cell Multiome
        ATAC + Gene Expression data among other data types. The main
        Bioconductor data representations used are SingleCellExperiment
        and RaggedExperiment.
biocViews: Software, Infrastructure, DataImport, SingleCell
VignetteBuilder: knitr
License: Artistic-2.0
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
BugReports: https://github.com/waldronlab/TENxIO/issues
URL: https://github.com/waldronlab/TENxIO
Collate: 'TENxFile-class.R' 'TENxFileList-class.R'
        'TENxFragments-class.R' 'TENxH5-class.R' 'TENxIO-package.R'
        'TENxMTX-class.R' 'TENxPeaks-class.R' 'TENxTSV-class.R'
        'utils.R'
Date: 2025-12-05
Config/pak/sysreqs: libssl-dev libx11-dev zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:59:13 UTC
RemoteUrl: https://github.com/bioc/TENxIO
RemoteRef: RELEASE_3_23
RemoteSha: 20544bc09f909783c1d3b581838275bafeed6eab
NeedsCompilation: no
Packaged: 2026-04-29 15:07:14 UTC; root
Author: Marcel Ramos [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3242-0582>),
  NCI [fnd] (GrantNo.: U24CA289073)
Maintainer: Marcel Ramos <marcel.ramos@sph.cuny.edu>
Built: R 4.6.0; ; 2026-04-29 15:52:33 UTC; unix

# Generated by roxygen2: do not edit by hand

export(TENxFile)
export(TENxFileList)
export(TENxFragments)
export(TENxH5)
export(TENxMTX)
export(TENxPeaks)
export(TENxTSV)
exportClasses(TENxFile)
exportClasses(TENxFileList)
exportClasses(TENxFragments)
exportClasses(TENxH5)
exportClasses(TENxMTX)
exportClasses(TENxPeaks)
exportClasses(TENxTSV)
exportMethods("annotation<-")
exportMethods(annotation)
exportMethods(decompress)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(genome)
exportMethods(import)
exportMethods(metadata)
exportMethods(path)
exportMethods(rowData)
exportMethods(rowRanges)
exportMethods(show)
import(SingleCellExperiment)
import(SummarizedExperiment)
importClassesFrom(BiocIO,BiocFile)
importFrom(BiocBaseUtils,checkInstalled)
importFrom(BiocBaseUtils,isScalarCharacter)
importFrom(BiocBaseUtils,selectSome)
importFrom(BiocBaseUtils,setSlots)
importFrom(BiocGenerics,"annotation<-")
importFrom(BiocGenerics,annotation)
importFrom(BiocGenerics,path)
importFrom(BiocIO,decompress)
importFrom(BiocIO,import)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,makeGRangesFromDataFrame)
importFrom(Matrix,readMM)
importFrom(MatrixGenerics,rowRanges)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,SimpleList)
importFrom(S4Vectors,isEmpty)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,splitAsList)
importFrom(Seqinfo,"genome<-")
importFrom(Seqinfo,genome)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,slotNames)
importFrom(readr,read_tsv)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,untar)

## subset H5 file
library(rhdf5)
# setwd("~/data/10x/pbmc_3k")
h5f <- "pbmc_granulocyte_sorted_3k_filtered_feature_bc_matrix.h5"
as.object_size(file.info(h5f)$size)
#' [1] "38.8 MB"

h5new <- "inst/extdata/pbmc_granulocyte_ff_bc_ex.h5"
newmat <- HDF5Array::TENxMatrix(h5f, "matrix")[1:10, 1:10]
HDF5Array::writeTENxMatrix(newmat, h5new, group = "matrix", verbose = TRUE)

tkeys <- h5read(h5f, "/matrix/features/_all_tag_keys") ## same
myidx <- list(1:10)
h1 <- h5read(h5f, "/matrix/features/feature_type", index = myidx)
h2 <- h5read(h5f, "/matrix/features/genome", index = myidx)
h3 <- h5read(h5f, "/matrix/features/id", index = myidx)
h4 <- h5read(h5f, "/matrix/features/interval", index = myidx)
h5 <- h5read(h5f, "/matrix/features/name", index = myidx)

h5createGroup(h5new, "/matrix/features/")
h5write(tkeys, h5new, "/matrix/features/_all_tag_keys")
h5write(h1, h5new, "/matrix/features/feature_type")
h5write(h2, h5new, "/matrix/features/genome")
h5write(h3, h5new, "/matrix/features/id")
h5write(h4, h5new, "/matrix/features/interval")
h5write(h5, h5new, "/matrix/features/name")

TENxH5(h5new)
as.object_size(file.info(h5new)$size, "kB")
#' [1] "30.2 kB"


h5f <- "10k_pbmc_ATACv2_nextgem_Chromium_Controller_filtered_peak_bc_matrix.h5"
h5ls(h5f)
newmat <- HDF5Array::TENxMatrix(h5f, "matrix")[1:10, 1:10]
h5new <- "inst/extdata/10k_pbmc_ATACv2_f_bc_ex.h5"
# file.remove(h5new)
HDF5Array::writeTENxMatrix(newmat, h5new, group = "matrix", verbose = TRUE)

tkeys <- h5read(h5f, "/matrix/features/_all_tag_keys") ## same
myidx <- list(1:10)
h1 <- h5read(h5f, "/matrix/features/feature_type", index = myidx)
h2 <- h5read(h5f, "/matrix/features/genome", index = myidx)
h3 <- h5read(h5f, "/matrix/features/id", index = myidx)
h4 <- h5read(h5f, "/matrix/features/derivation", index = myidx)
h5 <- h5read(h5f, "/matrix/features/name", index = myidx)

h5createGroup(h5new, "/matrix/features/")
h5write(tkeys, h5new, "/matrix/features/_all_tag_keys")
h5write(h1, h5new, "/matrix/features/feature_type")
h5write(h2, h5new, "/matrix/features/genome")
h5write(h3, h5new, "/matrix/features/id")
h5write(h4, h5new, "/matrix/features/derivation")
h5write(h5, h5new, "/matrix/features/name")

h5ls(h5new)
aa <- TENxH5(h5new, ranges = "/features/id")
import(aa)

as.object_size(file.size(h5new), "kB")
#' [1] "30.1 kB"


R.utils::gunzip(
    "filtered_feature_bc_matrix/matrix.mtx.gz",
    "filtered_feature_bc_matrix/test_matrix.mtx"
)
gg <- readLines("test_matrix.mtx", n = 13)
gg[[3]] <- "171 10 10"

writeLines(gg, "test_write_matrix.mtx")
assay(import(TENxMTX("test_write_matrix.mtx")))

file.copy("test_write_matrix.mtx", "inst/extdata/pbmc_3k_ff_bc_ex.mtx")

as.object_size(file.size(mtxf), unit = "kB")
#' [1] "0.2 kB"


# TENxFragments -----------------------------------------------------------

library(Rsamtools)
fr <- "pbmc_granulocyte_sorted_3k_atac_fragments.tsv"
tf <- TabixFile(fr, yieldSize = 10)
examp <- read.table(textConnection(scanTabix(tf)[[1]]))

names(examp) <- c("chrom", "chromStart", "chromEnd", "barcode", "readSupport")
readr::write_tsv(
    examp, "inst/extdata/pbmc_3k_atac_ex_fragments.tsv", col_names = FALSE
)
#' [1] "0.2 kB"
system2("bgzip inst/extdata/pbmc_3k_atac_ex_fragments.tsv", stdout = TRUE)
system("tabix -p bed inst/extdata/pbmc_3k_atac_ex_fragments.tsv.gz")

fragf <- "inst/extdata/pbmc_3k_atac_ex_fragments.tsv.gz"
as.object_size(file.size(fragf), unit = "kB")
#' [1] "0.2 kB"
as.object_size(file.size(paste0(fragf, ".tbi")), unit = "kB")
#' [1] "0.1 kB"

aa <- import(TENxFragments(fragf, yieldSize = 10))


# Peaks -------------------------------------------------------------------

fi <- "pbmc_granulocyte_sorted_3k_atac_peak_annotation.tsv"
pkt <- readr::read_tsv(fi, n_max = 10)
newpeak <- "inst/extdata/pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv"
readr::write_tsv(pkt, newpeak)
as.object_size(file.size(newpeak), unit = "kB")
#' [1] "0.5 kB"

import(TENxPeaks(newpeak))


# TENxFileList ------------------------------------------------------------

basedir <- "pbmc_granulocyte_sorted_3k_filtered_feature_bc_matrix/filtered_feature_bc_matrix/"
bc <- readr::read_tsv(
    file.path(basedir, "barcodes.tsv.gz"), col_names = FALSE, n_max = 10
)
readr::write_tsv(
    bc, file.path(basedir, "barcodes.tsv.gz"), col_names = FALSE
)
ft <- readr::read_tsv(
    file.path(basedir, "features.tsv.gz"), col_names = FALSE, n_max = 10
)
readr::write_tsv(
    ft, file.path(basedir, "features.tsv.gz"), col_names = FALSE
)
# file.copy("TENxIO/inst/extdata/pbmc_3k_ff_bc_ex.mtx",
#   file.path(basedir, "matrix.mtx"))
R.utils::gzip(
    file.path(basedir, "matrix.mtx"), file.path(basedir, "matrix.mtx.gz"),
    overwrite = TRUE
)
tar(
    "TENxIO/inst/extdata/pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
    "./pbmc_granulocyte_sorted_3k_filtered_feature_bc_matrix",
    compression = "gzip"
)

import(
    TENxFileList(
    "TENxIO/inst/extdata/pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz"
    )
)

as.object_size(
    file.size("pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz"), "KB"
)
#' [1] "0.7 kB"

#' TENxFile: General purpose class for 10X files
#'
#' @description The `TENxFile` class is the default representation for
#'   unrecognized subclasses. It inherits from the BiocFile class and adds a few
#'   additional slots. The constructor function can handle typical 10X file
#'   types. For more details, see the constructor function documentation.
#'
#' @slot extension character(1) The file extension as extracted from the file
#'   path or overridden via the `ext` argument in the constructor function.
#'
#' @slot colidx integer(1) The column index corresponding to the columns in the
#'   file that will subsequently be imported
#'
#' @slot rowidx integer(1) The row index corresponding to rows in the file that
#'   will subsequently be imported
#'
#' @slot remote logical(1) Whether the file exists on the web, i.e., the
#'   `resource` is a URL
#'
#' @slot compressed logical(1) Whether the file is compressed with, e.g., `.gz`
#'
#' @importClassesFrom BiocIO BiocFile
#' @importFrom BiocIO import
#' @importFrom BiocGenerics path
#' @importFrom S4Vectors mcols
#' @importFrom methods new is as
#'
#' @exportClass TENxFile
.TENxFile <- setClass(
    Class = "TENxFile",
    contains = "BiocFile",
    slots = c(
        extension = "character",
        colidx = "integer", rowidx = "integer",
        remote = "logical", compressed = "logical"
    )
)

.check_file_exists <- function(object) {
    op <- path(object)
    if (file.exists(op) || RCurl::url.exists(op))
        TRUE
    else
        "Path or URL to the file must be valid"
}

.validTENxFile <- function(object) {
    .check_file_exists(object)
}

S4Vectors::setValidity2("TENxFile", .validTENxFile)

# TENxFile constructor ----------------------------------------------------

.remove_query <- function(fname) {
    vapply(strsplit(fname, "\\?"), `[[`, character(1L), 1L)
}

.get_ext <- function(fname) {
    if (is(fname, "TENxFile"))
        ext <- fname@extension
    else {
        fname <- .remove_query(fname)
        split_files <- strsplit(basename(fname), "\\.")
        ext <- vapply(split_files, function(file) {
            paste0(
                utils::tail(file, -1),
                collapse = "."
            )
        }, character(1L))
    }
    if (identical(ext, ""))
        stop("No extension present, provide an 'extension' input")
    ext
}

#' TENxFile constructor function
#'
#' @description The `TENxFile` constructor function serves as the
#'   auto-recognizer function for 10X files. It can import several different
#'   file extensions, namely:
#' \preformatted{
#'     * H5 - on-disk HDF5
#'     * MTX - matrix market
#'     * .tar.gz - compressed tarball
#' }
#'
#' @details **Note** that the example below includes the use of a large ~ 4 GB
#'   `ExperimentHub` resource obtained from the 10X website.
#'
#' @param resource character(1) The path to the file
#'
#' @param extension character(1) The file extension for the given resource. It
#'   can usually be obtained from the file path. An override can be provided
#'   especially for `ExperimentHub` resources where the file extension is
#'   removed.
#'
#' @param ... Additional inputs to the low level class generator functions
#'
#' @return A subclass of `TENxFile` according to the input file extension
#'
#' @examples
#' if (interactive()) {
#'
#'     ## from ExperimentHub
#'     hub <- ExperimentHub::ExperimentHub()
#'     fname <- hub[["EH1039"]]
#'     TENxFile(fname, extension = "h5", group = "mm10", version = "2")
#'     TENxFile(fname, extension = "h5", group = "mm10", version = "2") |>
#'         metadata()
#'
#' }
#' @export
TENxFile <- function(resource, extension, ...) {
    if (missing(extension))
        extension <- .get_ext(resource)
    TENxFUN <- switch(
        extension,
        h5 = TENxH5,
        mtx = TENxMTX,
        mtx.gz = TENxMTX,
        tar.gz = TENxFileList,
        tsv.gz = TENxTSV,
        tsv = TENxTSV,
        .TENxFile
    )
    TENxFUN(resource = resource,  extension = extension, ...)
}

#' @describeIn TENxFile-class `metadata` method for `TENxFile` objects
#' 
#' @param x An object of class `TENxFile`, `TENxFileList`, `TENxMTX`, `TENxH5`,
#'   `TENxPeaks`, `TENxTSV`, or derivatives
#' 
#' @param ... Additional arguments (not used)
#'
#' @return A list of metadata for the given object
#'
#' @importFrom S4Vectors metadata
#' @importFrom methods slotNames
#'
#' @exportMethod metadata
setMethod("metadata", "TENxFile", function(x, ...) {
    sn <- slotNames(x)
    snl <- structure(sn, .Names = sn)
    metadata <- lapply(snl, getElement, object = x)
    metadata[["resource"]] <- basename(metadata[["resource"]])
    list(TENxFile = metadata)
})

expect_error(
    TENxFile(tempfile())
)

test_file <- tempfile()
file.create(test_file)
expect_error(
    tenxfile <- TENxFile(test_file)
)
tenxfile <- TENxFile(test_file, ".tsv")
expect_true(
    validObject(tenxfile)
)

file.remove(test_file)

test_file <- tempfile(fileext = ".tsv")
file.create(test_file)
expect_true(
    is(tenxfile <- TENxFile(test_file), "TENxFile")
)

expect_identical(
    tenxfile@resource, test_file
)

expect_identical(
    tenxfile@extension, "tsv"
)

file.remove(test_file)

test_file <- tempfile(fileext = ".bed")
file.create(test_file)

tenxfile <- TENxFile(test_file)

expect_identical(
    tenxfile@extension, "bed"
)

HDF5_conf <- system.file("include", "H5pubconf.h", package  = "Rhdf5lib")
configLine <-
    grepl(".*H5_HAVE_ROS3_VFD\\s+1", suppressWarnings(readLines(HDF5_conf)))

if (any(configLine)) {
    pbmc_url <- paste0(
        "https://raw.githubusercontent.com/waldronlab/TENxIO/",
        "devel/inst/extdata/10k_pbmc_ATACv2_f_bc_ex.h5"
    )
    
    remoteh5 <- TENxFile(pbmc_url)
    
    expect_true(
        is(remoteh5, "TENxH5")
    )
    
    expect_true(
        remoteh5@remote
    )
}

#' TENxFileList: A list-like representation for TENxFiles
#'
#' @description This class was designed to mainly handle tarballs from
#' 10X Genomics. The typical file extension for these tarballs is `.tar.gz`.
#'
#' @details
#' These tarballs usually contain three files:
#'   1. `matrix.mtx.gz` - the counts matrix
#'   2. `features.tsv.gz` - row metadata usually represented as `rowData`
#'   3. `barcodes.tsv.gz` - column names corresponding to cell barcode
#'   identifiers
#' Note that version '2' includes `genes.tsv.gz` instead of `features.tsv.gz` in
#' version '3'.
#' 
#' An additional `ref` argument can be provided when the file contains
#' multiple `feature_type` in the file or "Type" in the `rowData`. By default,
#' the first type reported in `table()` is set as the `mainExpName` in the
#' `SingleCellExperiment` object.
#'
#' @slot listData list() The data in list format
#'
#' @slot extension character() A vector of file extensions for each file
#'
#' @slot compressed logical(1) Whether the file is compressed as `.tar.gz`
#'
#' @slot version character(1) The version number of the tarball usually either
#'   '2' or '3'
#'
#' @return A `TENxFileList` class object
#'
#' @exportClass TENxFileList
.TENxFileList <- setClass(
    Class = "TENxFileList",
    contains = "SimpleList",
    slots = c(
        listData = "list",
        extension = "character",
        compressed = "logical",
        version = "character"
    )
)

.validTENxFileList <- function(object) {
    if (!object@compressed) {
        validClasses <- vapply(object, function(ifile) {
            is(ifile, "TENxFile")
        }, logical(1L))
        if (all(validClasses))
            TRUE
        else
            "Some files are not of class 'TENxFile'"
    }
}

file.list.map <- data.frame(
    Version = c("3", "2"),
    features = c("features.tsv", "genes.tsv")
)

S4Vectors::setValidity2("TENxFileList", .validTENxFileList)

# Constructor -------------------------------------------------------------

#' TENxFileList: Represent groups of files from 10X Genomic
#'
#' @description This constructor function is meant to handle `.tar.gz` tarball
#'   files from 10X Genomics.
#'
#' @details
#' These tarballs usually contain three files:
#'   1. `matrix.mtx.gz` - the counts matrix
#'   2. `features.tsv.gz` - row metadata usually represented as `rowData`
#'   3. `barcodes.tsv.gz` - column names corresponding to cell barcode
#'   identifiers
#' If all above files are in the tarball, the import method will provide a
#' `SingleCellExperiment`. Otherwise, a simple list of imported data is given.
#' Note that version "3" uses 'features.tsv.gz' and version "2" uses
#' 'genes.tsv.gz'. If known, indicate the `version` argument in the
#' `TENxFileList` constructor function.
#'
#' @param ... Typically, a file path to a tarball archive. Can be named
#'   arguments corresponding to file paths, or a named list of file paths.
#'
#' @param version character(1) The version in the tarball. See details.
#'
#' @param compressed logical(1) Whether or not the file provided is compressed,
#'   usually as `tar.gz` (default FALSE)
#'
#' @return Either a `SingleCellExperiment` or a list of imported data
#'
#' @examples
#'
#' fl <- system.file(
#'     "extdata", "pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
#'     package = "TENxIO", mustWork = TRUE
#' )
#'
#' ## Method 1 (tarball)
#' TENxFileList(fl)
#'
#' ## metadata before import
#' metadata(TENxFileList(fl))
#'
#' ## import() method
#' import(TENxFileList(fl))
#'
#' ## metadata after import
#' import(TENxFileList(fl)) |>
#'     metadata()
#'
#' ## untar to simulate folder output
#' dir.create(tdir <- tempfile())
#' untar(fl, exdir = tdir)
#'
#' ## Method 2 (folder)
#' TENxFileList(tdir)
#' import(TENxFileList(tdir))
#'
#' ## Method 3 (list of TENxFile objects)
#' files <- list.files(tdir, recursive = TRUE, full.names = TRUE)
#' names(files) <- basename(files)
#' filelist <- lapply(files, TENxFile)
#'
#' TENxFileList(filelist, compressed = FALSE)
#'
#' ## Method 4 (SimpleList)
#' TENxFileList(as(filelist, "SimpleList"), compressed = FALSE)
#'
#' ## Method 5 (named arguments)
#' TENxFileList(
#'     barcodes.tsv.gz = TENxFile(files[1]),
#'     features.tsv.gz = TENxFile(files[2]),
#'     matrix.mtx.gz = TENxFile(files[3])
#' )
#'
#' unlink(tdir, recursive = TRUE)
#'
#' @export
TENxFileList <- function(..., version, compressed = FALSE) {
    dots <- S4Vectors::SimpleList(...)
    exts <- dots[["extension"]]
    if (length(names(dots)))
        dots <- dots[names(dots) != "extension"]
    undots <- dots[[1L]]
    if (identical(length(dots), 1L)) {
        if (is.list(undots) || is(undots, "SimpleList")) {
            exts <- vapply(undots, .get_ext, character(1L))
            version <- .version_from_fnames(names(undots))
            dots <- undots
        }
        isdir <- try(file.info(undots)[["isdir"]], silent = TRUE)
        if (inherits(isdir, "try-error") && is(undots, "SimpleList")) {
            exts <- vapply(undots, .get_ext, character(1L))
            version <- .version_from_fnames(names(undots))
        } else if (isdir) {
            undots <- .get_files_from_folder(undots)
            dots <- lapply(undots, TENxFile)
        }
        if (is.character(undots) && is.null(exts))
            exts <- .get_ext(undots)
        if (missing(version) && is.character(undots) &&
                all(endsWith(undots, "tar.gz")))
            version <- .version_from_tarball(undots)
        else if (missing(version) && is.character(undots))
            version <- .version_from_fnames(undots)
        if (is(undots, "TENxFile")) {
            exts <- undots@extension
            version <- undots@version
        }
    } else {
        exts <- vapply(dots, .get_ext, character(1L))
        version <- .version_from_fnames(dots)
    }
    if (identical(exts, "tar.gz"))
        compressed <- TRUE
    .TENxFileList(
        dots, extension = exts, compressed = compressed, version = version
    )
}

.get_files_from_folder <- function(folder) {
    files <- list.files(folder, full.names = TRUE, recursive = TRUE)
    names(files) <- basename(files)
    files
}

#' @importFrom utils untar tail
.TENxUntar <- function(con) {
    dir.create(tempdir <- tempfile())
    untar(path(con), exdir = tempdir)
    tempdir
}

.get_path <- function(object) {
    if (is(object, "TENxFile"))
        path(object)
    else
        object
}

#' @describeIn TENxFileList-class Obtain file paths for all files in the object
#'   as a vector
#'
#' @inheritParams BiocGenerics::path
#'
#' @export
setMethod("path", "TENxFileList", function(object, ...) {
    vapply(object, .get_path, character(1L))
})

#' @describeIn TENxFileList-class An intermediate method for decompressing
#'   (via untar) the contents of a `.tar.gz` file list
#'
#' @importFrom BiocIO decompress
#' @importFrom BiocBaseUtils setSlots
#'
#' @inheritParams BiocIO::decompress
#'
#' @param manager A `ConnectionManager` internal instance; currently not used.
#'
#' @export
setMethod("decompress", "TENxFileList", function(manager, con, ...) {
    res_ext <- con@extension
    if (!identical(length(res_ext), 1L))
        stop("The 'extension' of 'con' must be from a single compressed file")
    if (is.na(res_ext))
        res_ext <- .get_ext(path(con))
    if (con@compressed) {
        if (identical(res_ext, "tar.gz")) {
            tenfolder <- .TENxUntar(con)
            gfolder <- list.files(tenfolder, full.names = TRUE)
            if (file.info(gfolder)$isdir)
                gfiles <- list.files(
                    gfolder, recursive = TRUE, full.names = TRUE
                )
            else
                gfiles <- gfolder
            newlistdata <- lapply(.setNames(gfiles, basename(gfiles)), TENxFile)
            con <- setSlots(
                object = con, listData = newlistdata, compressed = FALSE
            )
        } else {
            stop("Extension type: ", res_ext, " not supported")
        }
    }
    con
})

.version_from_fnames <- function(fnames) {
    if (is(fnames, "SimpleList"))
        fnames <- names(fnames)
    else
        fnames <- basename(fnames)
    if (any(grepl("features.tsv[\\.gz]*", fnames)))
        version <- "3"
    else if (any(grepl("genes.tsv[\\.gz]*", fnames)))
        version <- "2"
    else
        NA_character_
}

.version_from_tarball <- function(tarball) {
    flist <- untar(tarball, list = TRUE)
    .version_from_fnames(flist)
}

#' @describeIn TENxFileList-class Recursively import files within a
#'   `TENxFileList`
#'
#' @inheritParams BiocIO::import
#'
#' @importFrom S4Vectors metadata<-
#'
#' @export
setMethod("import", "TENxFileList", function(con, format, text, ...) {
    if (con@compressed)
        fdata <- decompress(con = con)
    else
        fdata <- con
    dots <- list(...)
    datalist <- lapply(fdata, import)
    features <-
        .selectByVersion(file.list.map, version = con@version, "features")
    features <- grep(features, names(fdata), fixed = TRUE, value = TRUE)
    matrix <- grep("matrix.mtx", names(fdata), fixed = TRUE, value = TRUE)
    barcodes <- grep("barcodes.tsv", names(fdata), fixed = TRUE, value = TRUE)
    if (con@version %in% c("2", "3")) {
        mat <- datalist[[matrix]]
        colnames(mat) <- unlist(
            datalist[[barcodes]], use.names = FALSE
        )
        feats <- datalist[[features]]
        sce <- as(mat, "SingleCellExperiment")
        metadata(sce) <- metadata(con)
        if (!is.null(feats) && ncol(feats))
            rownames(sce) <- feats[[1L]]
        if (!is.null(rownames(sce)) && anyDuplicated(rownames(sce))) {
            warning(
                "'rownames' in assay are not unique; using 'make.unique()'"
            )
            rownames(sce) <- make.unique(rownames(sce))
        }
        if ("Chr" %in% names(feats)) {
            feats[is.na(feats[["Chr"]]), "Chr"] <- "NA_character_:0"
            rr <- GenomicRanges::makeGRangesFromDataFrame(
                feats, keep.extra.columns = TRUE
            )
            names(rr) <- rownames(sce)
            rowRanges(sce) <- rr
            okrows <- seqnames(rr) != "NA_character_"
            sce <- sce[okrows, ]
            feats <- feats[as.logical(okrows), , drop = FALSE]
        } else {
            rowData(sce) <- feats
        }
        ref <- dots[["ref"]]
        types <- feats[["Type"]]
        if (!is.null(types) && length(types)) {
            if (!isScalarCharacter(ref))
                ref <- names(table(types)[1L])
            sce <- splitAltExps(sce, types, ref = ref)
        }
        sce
    } else {
        ## return a list for now
        datalist
    }
})

#' @describeIn TENxFileList-class `metadata` method for `TENxFileList` objects
#'
#' @param x An object of class `TENxFile`, `TENxFileList`, `TENxMTX`, `TENxH5`,
#'   `TENxPeaks`, `TENxTSV`, or derivatives
#' 
#' @param ... Additional arguments (not used)
#'
#' @exportMethod metadata
setMethod("metadata", "TENxFileList", function(x, ...) {
    sn <- slotNames(x)
    sn <- setdiff(sn, slotNames("SimpleList"))
    snl <- structure(sn, .Names = sn)
    metadata <- lapply(snl, getElement, object = x)
    ldata <- lapply(x@listData, function(y) {
        if (inherits(y, "TENxFile")) metadata(y) else basename(y)
    })
    metadata[["resources"]] <- ldata
    list(TENxFileList = metadata)
})

library(DropletTestFiles)
# listTestFiles()
library(ExperimentHub)
eh <- ExperimentHub()

path <- eh[["EH3686"]]
expect_warning(
    res <- TENxH5(
        resource = path, group = "mm10", ranges = NA_character_,
        version = "2", extension = "h5", rowidx = 1:10, colidx = 1:10
    ),
    pattern = "'group' not in known.*"
)
expect_true(
    is(res, "TENxH5")
)
expect_true(
    is(
        import(res, ref = NA),
        "SingleCellExperiment"
    )
)
expect_true(
    is(
        import(res),
        "SingleCellExperiment"
    )
)

path <- eh[["EH3688"]]
expect_error(
    TENxFileList(path, compressed = TRUE),
    "No extension present"
)
expect_true(
    is(
        res <- TENxFileList(
            path, extension = "tar.gz", compressed = TRUE, version = "2"
        ),
        "TENxFileList"
    )
)
expect_true(
    is(
        import(res),
        "SingleCellExperiment"
    )
)

fl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
    package = "TENxIO", mustWork = TRUE
)

txfl <- TENxFileList(fl)

expect_identical(
    length(txfl), 1L
)

expect_true(
    txfl@compressed
)

expect_identical(
    txfl@extension, "tar.gz"
)

fli <- import(txfl)

expect_true(
    is(fli, "SingleCellExperiment")
)

expect_true(
    is(assay(fli, i = "counts"), "Matrix")
)

expect_identical(
    dim(fli), c(10L, 10L)
)

expect_identical(
    names(rowData(fli)), c("ID", "Symbol", "Type")
)

expect_identical(
    mainExpName(fli), "Gene Expression"
)

expect_identical(
    assayNames(fli), "counts"
)

#' TENxPeaks: The class to represent 10x Peaks files
#'
#' This class is designed to work with the files denoted with "peak_annotation"
#' in the file name. These are usually produced as tab separated value files,
#' i.e., `.tsv`.
#'
#' @details This class is a straightforward class for handling peak data. It can
#'   be used in conjunction with the `annotation` method on a
#'   `SingleCellExperiment` to add peak information to the experiment. The
#'   ranged data is represented as a `GRanges` class object.
#'
#' @return A `TENxPeaks` class object
#'
#' @exportClass TENxPeaks
.TENxPeaks <- setClass(
    Class = "TENxPeaks",
    contains = "TENxFile"
)

.check_peaks <- function(object) {
    fpath <- path(object)
    if (!endsWith(fpath, "peak_annotation.tsv"))
        "Provide a 10X Peaks file ending in 'peak_annotation.tsv'"
    else
        TRUE
}

.validPeaksFile <- function(object) {
    .check_peaks(object)
}

S4Vectors::setValidity2("TENxPeaks", .validPeaksFile)

#' Import 10x peak annotation files from 10x
#'
#' This constructor function is designed to work with the files denoted with
#' "peak_annotation" in the file name. These are usually produced as tab
#' separated value files, i.e., `.tsv`.
#'
#' @details The output class allows handling of peak data. It can be used in
#'   conjunction with the `annotation` method on a `SingleCellExperiment` to add
#'   peak information to the experiment. The ranged data is represented as a
#'   `GRanges` class object.
#'
#' @inheritParams TENxFile
#'
#' @return A `GRanges` class object of peak locations
#'
#' @examples
#'
#' fi <- system.file(
#'     "extdata", "pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv",
#'     package = "TENxIO", mustWork = TRUE
#' )
#' peak_file <- TENxPeaks(fi)
#' peak_anno <- import(peak_file)
#' peak_anno
#'
#' example(TENxH5)
#'
#' ## Add peaks to an existing SCE
#' ## First, import the SCE from an example H5 file
#' h5f <- system.file(
#'     "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
#'     package = "TENxIO", mustWork = TRUE
#' )
#' con <- TENxH5(h5f)
#' sce <- import(con)
#' ## auto-import peaks when using annotation<-
#' annotation(sce, name = "peak_annotation") <- peak_file
#' annotation(sce)
#'
#' @export
TENxPeaks <- function(resource, extension, ...) {
    if (missing(extension))
        extension <- .get_ext(resource)
    .TENxPeaks(resource = resource, extension = extension, ...)
}

#' @describeIn TENxPeaks-class Import a peaks_annotation file from 10x as a
#'   `GRanges` representation
#'
#' @inheritParams BiocIO::import
#'
#' @export
setMethod("import", "TENxPeaks", function(con, format, ...) {
    if (identical(Sys.info()[["sysname"]], "Darwin"))
        readr::local_edition(1)
    panno <- readr::read_tsv(
        file = path(con), col_types = c("c", "n", "n", "c", "n", "c")
    )
    grs <- makeGRangesFromDataFrame(panno, keep.extra.columns = TRUE)
    metadata(grs) <- metadata(con)
    grs
})

#' @describeIn TENxPeaks-class Replacement method to add annotation data to a
#'   `SingleCellExperiment`
#'
#' @inheritParams BiocGenerics::annotation
#'
#' @importFrom S4Vectors metadata metadata<-
#' @importFrom BiocGenerics annotation<-
#'
#' @export
setReplaceMethod("annotation", "SingleCellExperiment",
    function(object, ..., value) {
        if (!is(value, "TENxPeaks"))
            stop("'value' must be of class 'TENxPeaks'")
        args <- list(...)
        append <- isTRUE(args[["append"]])
        anno_name <- args[["name"]]
        if (is.null(anno_name))
            anno_name <- "peak_annotation"
        meta <- metadata(object)
        annotated <- "annotation" %in% names(meta)
        if (annotated && !append)
            warning(
                "'append = FALSE'; replacing annotation in metadata",
                call. = FALSE
            )
        vlist <- structure(
            list(import(value)), .Names = anno_name
        )
        if (append)
            vlist <- append(meta[["annotation"]], vlist)
        metadata(object)[["annotation"]] <- vlist
        object
    }
)

#' @describeIn TENxPeaks-class Extraction method to obtain annotation data from
#'   a `SingleCellExperiment` representation
#'
#' @importFrom BiocGenerics annotation
#'
#' @export
setMethod("annotation", "SingleCellExperiment", function(object, ...) {
    metadata(object)[["annotation"]]
})

fi <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv",
    package = "TENxIO", mustWork = TRUE
)

peak_file <- TENxPeaks(fi)

expect_true(
    is(peak_file, "TENxPeaks")
)

expect_identical(
    peak_file@extension, "tsv"
)

peak_anno <- import(peak_file)

expect_true(
    is(peak_anno, "GRanges")
)

expect_identical(
    length(peak_anno), 10L
)

expect_identical(
    colnames(mcols(peak_anno)), c("gene", "distance", "peak_type")
)

sce <- SingleCellExperiment::SingleCellExperiment(
    assays = SimpleList(counts = matrix(rnorm(100), 10, 10))
)

expect_error(
    annotation(sce, name = "peak_annotation") <- peak_anno
)

expect_silent(
    annotation(sce, name = "peak_annotation") <- peak_file
)

expect_identical(
    metadata(sce), list(annotation = list(peak_annotation = peak_anno))
)

expect_identical(
    names(annotation(sce)), "peak_annotation"
)

expect_warning(
    annotation(sce, append = FALSE) <- peak_file
)

expect_silent(
    annotation(sce, append = TRUE) <- peak_file
)

expect_identical(
    length(metadata(sce)), 1L
)

expect_identical(
    length(metadata(sce)[["annotation"]]), 2L
)

expect_identical(
    names(metadata(sce)[["annotation"]]),
    c("peak_annotation", "peak_annotation")
)

#' TENxIO: A Bioconductor package for importing 10X Genomics files
#'
#' The package provides file classes based on `BiocIO` for common file
#' extensions found in the 10X Genomics website.
#'
#' @section Supported file types:
#' Here is a table of supported file and file extensions and their imported
#' classes:
#'
#' | **Extension**       | **Class**     | **Imported as**      |
#' |---------------------|---------------|----------------------|
#' | .h5                 | TENxH5        | SingleCellExperiment w/ TENxMatrix |
#' | .mtx / .mtx.gz      | TENxMTX       | SummarizedExperiment w/ dgCMatrix |
#' | .tar.gz             | TENxFileList  | SingleCellExperiment w/ dgCMatrix |
#' | peak_annotation.tsv | TENxPeaks     | GRanges              |
#' | fragments.tsv.gz    | TENxFragments | RaggedExperiment     |
#' | .tsv / .tsv.gz      | TENxTSV       | tibble               |
#'
#' @import SummarizedExperiment SingleCellExperiment
#'
#' @docType package
#'
#' @aliases TENxIO-package
#'
#' @name TENxIO
#'
"_PACKAGE"

#' @include TENxFile-class.R
#'
NULL

#' TENxFragments: A class to represent fragments data as `GRanges`
#'
#' This class is designed to work mainly with `fragments.tsv.gz` files from
#' 10x pipelines.
#'
#' @details Fragments data from 10x can be quite large. In order to speed up
#' the initial exploration of the data, we use a default of **200** records
#' for loading. Users can change this default value by specifying a new one
#' via the `yieldSize` argument in the constructor function.
#'
#' @slot which GRanges() A GRanges indicating the regions of interest. This
#'   get sent to `RSamtools` as the `param` input.
#'
#' @slot yieldSize numeric() The number of records to read by default, 200
#'   records will be imported. A warning will be emitted if not modified.
#'
#' @return A `TENxFragments` class object
#'
#' @exportClass TENxFragments
.TENxFragments <- setClass(
    Class = "TENxFragments",
    contains = "TENxFile",
    slots = c(which = "GenomicRanges", yieldSize = "numeric")
)

.check_fragments <- function(object) {
    fpath <- path(object)
    fragex <- endsWith(fpath, c("_fragments.tsv.gz", "_fragments.tsv"))
    if (!any(fragex))
        "Provide a 10X fragments file ending in '_fragments.tsv*'"
    else
        TRUE
}

.validTENxFragments <- function(object) {
    .check_fragments(object)
}

S4Vectors::setValidity2("TENxFragments", .validTENxFragments)

#' TENxFragments: Import fragments files from 10X
#'
#' @param resource character(1) The file path to the fragments resource, usually
#'   a compressed tabix file with extension `.tsv.gz`.
#'
#' @param which GRanges() A GRanges indicating the regions of interest. This
#'   get sent to `RSamtools` as the `param` input.
#'
#' @param yieldSize numeric() The number of records to read by default, 200
#'   records will be imported. A warning will be emitted if not modified.
#'
#' @param ... Further arguments to the class generator function (currently not
#'   used)
#'
#' @importFrom GenomicRanges GRanges makeGRangesFromDataFrame
#' @importFrom S4Vectors splitAsList mcols
#'
#' @return A `RaggedExperiment` object class
#'
#' @examples
#'
#' fr <- system.file(
#'     "extdata", "pbmc_3k_atac_ex_fragments.tsv.gz",
#'     package = "TENxIO", mustWork = TRUE
#' )
#'
#' tfr <- TENxFragments(fr)
#'
#' fra <- import(tfr)
#'
#' @export
TENxFragments <- function(resource, yieldSize = 200, which = GRanges(), ...) {
    if (missing(yieldSize) && missing(which))
        warning("Using default 'yieldSize' parameter")
    else if (!missing(which))
        yieldSize <- NA_integer_
    if (!is(which, "GRanges"))
        stop("'which' input must be 'GenomicRanges'")
    .TENxFragments(
        resource = resource, yieldSize = yieldSize, which = which,
        ...
    )
}

#' @describeIn TENxFragments Import method for representing fragments.tsv.gz
#'   data from 10x via `Rsamtools` and `RaggedExperiment`
#'
#' @importFrom utils read.table
#' @importFrom BiocBaseUtils checkInstalled
#'
#' @inheritParams BiocIO::import
#'
#' @export
setMethod("import", "TENxFragments", function(con, format, text, ...) {
    checkInstalled(c("Rsamtools", "RaggedExperiment"))
    which <- con@which
    yieldSize <- con@yieldSize
    tb <- Rsamtools::TabixFile(path(con), yieldSize = yieldSize)
    ex <- read.table(
        textConnection(Rsamtools::scanTabix(tb, param = which)[[1]])
    )
    ## https://support.10xgenomics.com/single-cell-atac/
    ## software/pipelines/latest/output/fragments
    names(ex) <- c("chrom", "chromStart", "chromEnd", "barcode", "readSupport")
    ggr <- makeGRangesFromDataFrame(
        ex, keep.extra.columns = TRUE, starts.in.df.are.0based = TRUE
    )
    grs <- splitAsList(ggr, mcols(ggr)$barcode)
    ggrl <- as(grs, "GRangesList")
    RaggedExperiment::RaggedExperiment(ggrl, metadata = metadata(con))
})

fr <- system.file(
    "extdata", "pbmc_3k_atac_ex_fragments.tsv.gz",
    package = "TENxIO", mustWork = TRUE
)
## yieldSize warning
expect_warning(
    tfr <- TENxFragments(fr)
)

tfr <- TENxFragments(fr, extension = "tsv.gz")
expect_identical(
    tfr@extension, "tsv.gz"
)
expect_silent(
    tfr <- TENxFragments(fr, yieldSize = 250)
)

expect_true(
    is(tfr, "TENxFragments")
)

fra <- import(tfr)
expect_true(
    is(fra, "RaggedExperiment")
)

expect_identical(
    assayNames(fra), c("barcode", "readSupport")
)

expect_identical(
    dim(fra), c(10L, 10L)
)

#' TENxTSV: A class to represent 10x tab separated values files
#'
#' @aliases TENxTSV
#'
#' @description This class is general purpose for reading in tabular data from
#'   the 10x Genomics website with the `.tsv` file extension. The class also
#'   supports compressed files, i.e., those with the `.tsv.gz` extension.
#'
#' @details Typical `.tsv` files obtained from the 10X website are compressed
#'   and contain information relevant to 'barcodes' and 'features'. Currently,
#'   the code only supports files such as `features.tsv.*` and `barcodes.tsv.*`.
#'
#' @inheritParams TENxMTX-class
#'
#' @return A `TENxTSV` class object; a `tibble` for the import method
#'
#' @exportClass TENxTSV
.TENxTSV <- setClass(Class = "TENxTSV", contains = "TENxFile")

#' @describeIn TENxTSV-class General import method for `tsv` files from 10x;
#'   using `readr::read_tsv` and returning a `tibble` representation
#'
#' @inheritParams BiocIO::import
#'
#' @importFrom readr read_tsv
#'
#' @export
setMethod("import", "TENxTSV", function(con, format, text, ...) {
    resource <- path(con)
    if (identical(Sys.info()[["sysname"]], "Darwin"))
        readr::local_edition(1)
    df <- readr::read_tsv(
        resource, col_names = FALSE, show_col_types = FALSE, progress = FALSE,
        ...
    )
    fname <- basename(resource)
    if (startsWith(fname, "features.tsv"))
        names(df) <- c("ID", "Symbol", "Type", "Chr", "Start", "End")
    else if (startsWith(fname, "barcodes.tsv"))
        names(df) <- "barcode"
    attr(df, "metadata") <- metadata(con)
    df
})

#' @rdname TENxTSV-class
#'
#' @inheritParams TENxMTX
#'
#' @export
TENxTSV <-  function(resource, compressed = FALSE, ...) {
    dots <- list(...)
    ext <- dots[["extension"]]
    if (is.null(ext))
        ext <- .get_ext(resource)
    compr <- identical(ext, "tsv.gz")
    if (!ext %in% c("tsv.gz", "tsv"))
        warning("File extension is not 'tsv'; import may fail", call. = FALSE)
    .TENxTSV(resource = resource, compressed = compr, extension = ext)
}

#' @describeIn TENxTSV-class `metadata` method for `TENxTSV` objects
#'
#' @param x A `TENxTSV` object
#'
#' @exportMethod metadata
setMethod("metadata", "TENxTSV", function(x, ...) {
    sn <- slotNames(x)
    snl <- structure(sn, .Names = sn)
    metadata <- lapply(snl, getElement, object = x)
    metadata[["resource"]] <- basename(metadata[["resource"]])
    list(TENxTSV = metadata)
})

#' TENxH5: The HDF5 file representation class for 10X Data
#'
#' @description This class is designed to work with 10x Single Cell datasets.
#'   It was developed using the PBMC 3k 10X dataset from the CellRanger v2
#'   pipeline.
#'
#' @slot version character(1) There are currently two recognized versions
#'   associated with 10X data, either version "2" or "3". See details for more
#'   information.
#'
#' @slot group character(1) The HDF5 group embedded within the file structure,
#'   this is usually either the "matrix" or "outs" group but other groups are
#'   supported as well.
#'
#' @slot ranges character(1) The HDF5 internal folder location embedded within
#'   the file that points to the ranged data information, e.g.,
#'   "/features/interval".
#'
#' @details The data version "3" mainly includes a "matrix" group and "interval"
#'   information within the file. Version "2" data does not include
#'   ranged-based information and has a different directory structure compared
#'   to version "3". See the internal `data.frame`: `TENxIO:::h5.version.map` for
#'   a map of fields and their corresponding file locations within the H5 file.
#'   This map is used to create the `rowData` structure from the file.
#'
#' @section import:
#'   The `import` method uses `DelayedArray::TENxMatrix` to represent matrix
#'   data. Generally, version 3 datasets contain associated genomic coordinates.
#'   The associated feature data, as displayed by the `rowData` method, is
#'   queried for the "Type" column which will indicate that a `splitAltExps`
#'   operation is appropriate. If a `ref` input is provided to the constructor
#'   function `TENxH5`, it will be used as the main experiment; otherwise, the
#'   most frequent category in the "Type" column will be used. For example,
#'   the Multiome ATAC + Gene Expression feature data contains both 'Gene
#'   Expression' and 'Peaks' labels in the "Type" column.
#'
#' @return A `TENxH5` class object
#'
#' @seealso [TENxH5]
#'
#' @include TENxFile-class.R
#'
#' @exportClass TENxH5
.TENxH5 <- setClass(
    Class = "TENxH5",
    contains = "TENxFile",
    slots = c(version = "character", group = "character", ranges = "character")
)

.get_h5_group <- function(fpath, remote) {
    l1 <- rhdf5::h5ls(fpath, recursive = FALSE, s3 = remote)
    l1[l1$otype == "H5I_GROUP", "name"]
}

.KNOWN_H5_GROUPS <- c("matrix", "outs")
.KNOWN_VERSIONS <- c("3", "2")

.check_h5_group <- function(group) {
    g_msg <- paste(.KNOWN_H5_GROUPS, collapse = ", ")
    if (!group %in% .KNOWN_H5_GROUPS)
        warning("'group' not in known 10X groups: ", g_msg, call. = FALSE)
}

.getDim <- function(file, group, remote) {
    rhdf5::h5read(file, paste0(group, "/", "shape"), s3 = remote)
}

.get_tenx_version <- function(group) {
    .KNOWN_VERSIONS[match(group, .KNOWN_H5_GROUPS)]
}

# Constructor -------------------------------------------------------------

#' TENxH5: Represent H5 files from 10X
#'
#' This constructor function was developed using the PBMC 3K dataset from 10X
#' Genomics (version 3). Other versions are supported and input arguments
#' `version` and `group` can be overridden.
#'
#' @details The various `TENxH5` methods including `rowData` and `rowRanges`,
#'   provide a snapshot of the data using a length 12 head and tail subset for
#'   efficiency. In contrast, methods such as `dimnames` and `dim` give
#'   a full view of the dimensions of the data. The `show` method provides
#'   relevant information regarding the dimensions of the data including
#'   metadata such as `rowData` and "Type" column, if available. The data files
#'   are `import`ed as `SingleCellExperiment` class objects.
#'
#'   An additional `ref` argument can be provided when the file contains
#'   multiple `feature_type` in the file or "Type" in the `rowData`. By default,
#'   the first type reported in `table()` is set as the `mainExpName` in the
#'   `SingleCellExperiment` object.
#'
#'   For data that do not contain genomic coordinate information, the `TENxH5`
#'   will fail to read `"/features/interval"` and will set the `ranges`
#'   argument to `NA_character_`.
#'
#'   The data version "3" mainly includes a "matrix" group and "interval"
#'   information within the file. Version "2" data does not include
#'   ranged-based information and has a different directory structure compared
#'   to version "3". See the internal `data.frame`: `TENxIO:::h5.version.map`
#'   for a map of fields and their corresponding file locations within the H5
#'   file. This map is used to create the `rowData` structure from the file.
#'
#' @inheritParams TENxFile
#'
#' @param version character(1) There are currently two recognized versions
#'   associated with 10X data, either version "2" or "3". See details for more
#'   information.
#'
#' @param group character(1) The HDF5 group embedded within the file structure,
#'   this is usually either the "matrix" or "outs" group but other groups are
#'   supported as well (e.g., "mm10").
#'
#' @param ranges character(1) The HDF5 internal folder location embedded within
#'   the file that points to the ranged data information, e.g.,
#'   "/features/interval". Set to `NA_character_` if range information is not
#'   present.
#'
#' @param rowidx,colidx numeric() A vector of indices corresponding to either
#'   rows or columns that will dictate the data imported from the file. The
#'   indices will be passed on to the `[` method of the `TENxMatrix`
#'   representation.
#'
#' @return Usually, a `SingleCellExperiment` instance
#'
#' @seealso `import` section in [TENxH5-class]
#'
#' @examples
#'
#' h5f <- system.file(
#'     "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
#'     package = "TENxIO", mustWork = TRUE
#' )
#'
#' TENxH5(h5f)
#'
#' import(TENxH5(h5f))
#'
#' h5f <- system.file(
#'     "extdata", "10k_pbmc_ATACv2_f_bc_ex.h5",
#'     package = "TENxIO", mustWork = TRUE
#' )
#'
#' ## Optional ref input, most frequent Type used by default
#' th5 <- TENxH5(h5f, ranges = "/features/id", ref = "Peaks")
#' th5
#' TENxH5(h5f, ranges = "/features/id")
#' import(th5)
#'
#' @export
TENxH5 <-
    function(resource, version, group, ranges, rowidx, colidx, ...)
{
    remote <- R.utils::isUrl(resource)
    group <- .get_h5_group(resource, remote)
    .check_h5_group(group)
    if (missing(version))
        version <- .get_tenx_version(group)
    dims <- .getDim(resource, group, remote)
    ext <- list(...)[["extension"]]
    if (is.null(ext))
        ext <- .get_ext(resource)
    if (!identical(tolower(ext), "h5"))
        warning("File extension is not 'h5'; import may fail", call. = FALSE)
    if (missing(ranges))
        ranges <- .selectByVersion(h5.version.map, version, "Ranges")
    ranges <- .validateRanges(resource, group, ranges, remote)
    if (missing(rowidx))
        rowidx <- seq_len(dims[[1L]])
    if (missing(colidx))
        colidx <- seq_len(dims[[2L]])
    .TENxH5(
        resource = resource, group = group, version = version, ranges = ranges,
        rowidx = rowidx,
        colidx = colidx,
        remote = remote,
        extension = ext
    )
}

h5.version.map <- data.frame(
    Version = c("3", "2"),
    ID = c("/features/id", "/genes"),
    Symbol = c("/features/name", "/gene_names"),
    Type = c("/features/feature_type", NA_character_),
    Ranges = c("/features/interval", NA_character_)
)

.selectByVersion <-
    function(df, version, select = !names(df) %in% c("Version", "Ranges"))
{
    if (missing(version) || is.na(version))
        NA_character_
    else
        df[df[["Version"]] == version, select]
}

.validateRanges <- function(file, group, ranges, remote) {
    if (!isScalarCharacter(ranges))
        NA_character_
    else
        tryCatch({
            rhdf5::h5read(
                file,
                name = paste0(group, ranges),
                index = list(1L),
                s3 = remote
            )
            ranges
        }, error = function(e) {
            warning(
                simpleWarning(
                    message = conditionMessage(e), call = conditionCall(e)
                )
            )
            NA_character_
        })
}

.read_rowData <- function(x, nrows) {
    gm <- .selectByVersion(h5.version.map, x@version)
    remote <- x@remote
    gm[] <- Filter(Negate(is.na), gm)
    res <- lapply(gm, function(colval) {
        readname <- paste0(x@group, colval)
        as.character(
            rhdf5::h5read(path(x), index = list(nrows), readname, s3 = remote)
        )
    })
    DF <- as(res, "DataFrame")
    if ("Type" %in% names(DF))
        DF[["Type"]] <- as.factor(DF[["Type"]])
    rownames(DF) <- nrows
    DF
}

#' @describeIn TENxH5 Generate the rowData ad hoc from a TENxH5 file
#'
#' @param x A `TENxH5` object
#'
#' @inheritParams SummarizedExperiment::rowData
#'
#' @export
setMethod("rowData", "TENxH5", function(x, use.names = TRUE, ...) {
    nrows <- list(...)[["rows"]]
    ## Implement a smaller index for display purposes only
    mxrow <- max(x@rowidx)
    if (is.null(nrows) && mxrow > 12)
        nrows <- c(seq(6), mxrow - 5:0)
    .read_rowData(x, nrows)
})

#' @describeIn TENxH5 Get the dimensions of the data as stored in the file
#' @export
setMethod("dim", "TENxH5", function(x) {
    c(length(x@rowidx), length(x@colidx))
})

#' @describeIn TENxH5 Get the dimension names from the file
#' @export
setMethod("dimnames", "TENxH5", function(x) {
    id <- .selectByVersion(h5.version.map, x@version, "ID")
    list(
        rhdf5::h5read(path(x), paste0(x@group, "/", id), s3 = x@remote),
        rhdf5::h5read(path(x), paste0(x@group, "/", "barcodes"), s3 = x@remote)
    )
})

#' @describeIn TENxH5 Read genome string from file
#' @importFrom Seqinfo genome genome<-
#' @export
setMethod("genome", "TENxH5", function(x) {
    group <- x@group
    version <- x@version
    remote <- x@remote
    if (is.na(x@ranges))
        stop("'rowRanges' data not available, e.g., in '/features/interval'")
    gens <- rhdf5::h5read(
        path(x), paste0(group, "/", "features/genome"), s3 = remote
    )
    ugens <- unique(gens)
    intervals <- rhdf5::h5read(
        path(x), paste0(group, "/", x@ranges), s3 = remote
    )
    splitints <- strsplit(intervals, ":", fixed = TRUE)
    seqnames <- vapply(splitints, `[[`, character(1L), 1L)
    if (any(seqnames == "NA"))
        warning("'seqlevels' contain NA values")
    if (identical(length(ugens), 1L)) {
        useq <- unique(seqnames)
        .setNames(rep(ugens, length(useq)), useq)
    } else
        vapply(split(gens, seqnames), unique, character(1L))
})

#' @describeIn TENxH5 Read interval data and represent as GRanges
#' @importFrom S4Vectors mcols<-
#' @export
setMethod("rowRanges", "TENxH5", function(x, ...) {
    group <- x@group
    version <- x@version
    remote <- x@remote
    if (is.na(x@ranges))
        stop("'rowRanges' data not available, e.g., in '/features/interval'")
    rows <- list(...)[["rows"]]
    ## Implement a smaller index for display purposes only
    mxrow <- max(x@rowidx)
    if (is.null(rows) && mxrow > 12)
        rows <- c(seq(6), mxrow - 5:0)
    interval <- rhdf5::h5read(
        path(x), paste0(group, x@ranges), list(rows), s3 = remote
    )
    ## Hack to allow NA ranges for later removal (keeping data parallel)
    interval[interval == "NA"] <- "NA_character_:0"
    gr <- as(as.character(interval), "GRanges")
    names(gr) <- rows
    mcols(gr) <- rowData(x, rows = rows)
    genome(gr) <- genome(x)
    cat("preview <= 12 rowRanges:", basename(x@resource), "\n")
    gr
})

#' @describeIn TENxH5 Import TENxH5 data as a SingleCellExperiment; see section
#'   below
#'
#' @importFrom MatrixGenerics rowRanges
#' @importFrom BiocBaseUtils isScalarCharacter
#' @importFrom S4Vectors isEmpty
#'
#' @inheritParams BiocIO::import
#'
#' @export
setMethod("import", "TENxH5", function(con, format, text, ...) {
    matrixdata <-
        HDF5Array::TENxMatrix(path(con), con@group)[con@rowidx, con@colidx]
    dots <- list(...)
    if (!con@version %in% c("2", "3"))
        stop("Version not supported.")
    sce <- SingleCellExperiment(
        assays = list(counts = matrixdata),
        metadata = metadata(con)
    )
    if (anyDuplicated(rownames(sce))) {
        warning(
            "'rownames' in assay are not unique; using 'make.unique()'"
        )
        rownames(sce) <- make.unique(rownames(sce))
    }
    if (identical(con@version, "3")) {
        if (!is.na(con@ranges)) {
            rr <- rowRanges(con, rows = con@rowidx)
            names(rr) <- mcols(rr)[["ID"]]
            rowRanges(sce) <- rr
            okrows <- seqnames(rr) != "NA_character_"
            ## remove stand-in NA values
            sce <- sce[okrows, ]
            con@rowidx <- con@rowidx[as.logical(okrows)]
        }
        if (!isEmpty(rowData(con))) {
            rowData(sce) <- .read_rowData(con, con@rowidx)
        }
    }
    types <- rowData(sce)[["Type"]]
    if (!is.null(types)) {
        ref <- dots[["ref"]]
        if (is.null(ref) || is.na(ref))
            ref <- names(table(types)[1L])
        if (isScalarCharacter(ref))
            sce <- splitAltExps(sce, types, ref = ref)
    }
    sce
})

#' @describeIn TENxH5 Display a snapshot of the contents within a TENxH5 file
#'   before import
#'
#' @param object A `TENxH5` class object
#'
#' @importFrom BiocBaseUtils selectSome
#' @export
setMethod("show", "TENxH5", function(object) {
    rn <- cn <- NULL
    rno <- rownames(object)
    cno <- colnames(object)
    if (length(rno))
        rn <- selectSome(rno)
    if (length(cno))
        cn <- selectSome(cno)
    rd <- rowData(object)
    rdnames <- selectSome(names(rd))
    rdnamecount <- paste0("names(", length(rd), "):")

    cat(
        class(object), "object",
        "\nresource:", path(object),
        "\ndim:", dim(object),
        "\nrownames:", rn,
        "\nrowData", rdnamecount, rdnames,
        if (length(rd[["Type"]]))
            "\n  Type:", selectSome(levels(rd[["Type"]])),
        "\ncolnames:", cn,
        "\n", sep = " "
    )
})

h5f <- system.file(
    "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)
tenxh5 <- TENxH5(h5f)
expect_true(
    is(tenxh5, "TENxH5")
)

expect_identical(
    genome(tenxh5),
    c(chr1 = "GRCh38")
)

# set to NA_character_ to test
tenxh5@ranges <- NA_character_
expect_error(
    genome(tenxh5)
)
expect_error(
    rowRanges(tenxh5)
)
tenxh5@ranges <- "/features/interval"

sceh5 <- import(tenxh5)
expect_true(
    is(sceh5, "SingleCellExperiment")
)

expect_identical(
    mainExpName(sceh5), "Gene Expression"
)

expect_identical(
    colnames(rowData(sceh5)), c("ID", "Symbol", "Type")
)

expect_identical(
    assayNames(sceh5), "counts"
)

expect_identical(
    dim(sceh5), c(10L, 10L)
)

#' TENxMTX: The Matrix Market representation class for 10X Data
#'
#' @description This class is designed to work with 10x MTX datasets,
#'   particularly from the multiome pipelines.
#'
#' @slot compressed logical(1) Whether or not the file is in compressed format,
#'   usually gzipped (`.gz`).
#'
#' @details The `TENxMTX` class is a straightforward implementation that allows
#'   the user to import a Matrix Market file format using `Matrix::readMM`.
#'   Currently, it returns a `SummarizedExperiment` with an internal `dgCMatrix`
#'   assay. To request other formats, please open an issue on GitHub.
#'
#' @return A `TENxMTX` class object
#'
#' @exportClass TENxMTX
.TENxMTX <- setClass(Class = "TENxMTX", contains = "TENxFile")

#' TENxMTX: Represent Matrix Market Format Files from 10X
#'
#' This constructor function accepts `.mtx` and `.mtx.gz` compressed formats
#' for eventual importing. It is mainly used with tarball files from 10X
#' Genomics, where more annotation data is included. Importing solely the
#' `.mtx` format will provide users with a `SummarizedExperiment` with an assay
#' of class `dgCMatrix` from the `Matrix` package. Currently, other formats are
#' not supported but if you'd like to request support for a format, please open
#' an issue on GitHub.
#'
#' @inheritParams TENxFile
#'
#' @param compressed logical(1) Whether the resource file is compressed (default
#'   FALSE)
#'
#' @return A `SummarizedExperiment` instance with a `dgCMatrix` in the assay
#'
#' @examples
#'
#' mtxf <- system.file(
#'     "extdata", "pbmc_3k_ff_bc_ex.mtx",
#'     package = "TENxIO", mustWork = TRUE
#' )
#'
#' con <- TENxMTX(mtxf)
#'
#' import(con)
#'
#' @export
TENxMTX <- function(resource, compressed = FALSE, ...) {
    dots <- list(...)
    ext <- dots[["extension"]]
    if (is.null(ext))
        ext <- .get_ext(resource)
    compr <- identical(ext, "mtx.gz")
    if (!ext %in% c("mtx.gz", "mtx"))
        warning("File extension is not 'mtx'; import may fail", call. = FALSE)
    .TENxMTX(resource = resource, compressed = compr, extension = ext)
}

#' @describeIn TENxMTX Import method mainly for mtx.gz files from 10x
#'
#' @importFrom S4Vectors SimpleList
#' @importFrom Matrix readMM
#'
#' @inheritParams BiocIO::import
#'
#' @export
setMethod("import", "TENxMTX", function(con, format, text, ...) {
    mtxf <- Matrix::readMM(path(con))
    ## coerce to common use class
    mtxf <- as(mtxf, "CsparseMatrix")
    SummarizedExperiment(
        assays = S4Vectors::SimpleList(counts = mtxf),
        metadata = metadata(con)
    )
})

mtxf <- system.file(
    "extdata", "pbmc_3k_ff_bc_ex.mtx",
    package = "TENxIO", mustWork = TRUE
)

con <- TENxMTX(mtxf)

expect_true(
    is(con, "TENxMTX")
)

expect_identical(
    con@extension, "mtx"
)

expect_false(
    con@compressed
)

mtx <- import(con)
expect_true(
    is(mtx, "SummarizedExperiment")
)

expect_identical(
    assayNames(mtx), "counts"
)

expect_identical(
    dim(mtx), c(171L, 10L)
)

---
title: "TENxIO: Import Single Cell Data Files"
author: "Marcel Ramos"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{TENxIO Quick Start Guide}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    number_sections: no
    toc: yes
    toc_depth: 4
package: TENxIO
---

<!-- badges: start -->
<!-- badges: end -->

```{r, setup, include=FALSE}
knitr::opts_chunk$set(cache = TRUE)
```

# Introduction

`TENxIO` allows users to import 10X pipeline files into known Bioconductor
classes. The package is not comprehensive, there are file types that are not
supported. For Visium datasets, we direct users to the `VisiumIO` package on
Bioconductor. TENxIO consolidates functionality from `DropletUtils`. If you
would like a file format to be supported, open an issue at
https://github.com/waldronlab/TENxIO.

# Supported Formats

| **Extension**       | **Class**     | **Imported as**      |
|---------------------|---------------|----------------------|
| .h5                 | TENxH5        | SingleCellExperiment w/ TENxMatrix |
| .mtx / .mtx.gz      | TENxMTX       | SummarizedExperiment w/ dgCMatrix |
| .tar.gz             | TENxFileList  | SingleCellExperiment w/ dgCMatrix |
| peak_annotation.tsv | TENxPeaks     | GRanges              |
| fragments.tsv.gz    | TENxFragments | RaggedExperiment     |
| .tsv / .tsv.gz      | TENxTSV       | tibble               |
| spatial.tar.gz      | TENxSpatialList | inter. DataFrame list |

# Tested 10X Products

We have tested these functions with _some_
[datasets](https://www.10xgenomics.com/resources/datasets) from 10x Genomics
including those from:

* Single Cell Gene Expression
* Single Cell ATAC
* Single Cell Multiome ATAC + Gene Expression
* Spatial Gene Expression

Note. That extensive testing has not been performed and the codebase may require
some adaptation to ensure compatibility with all pipeline outputs.

# Bioconductor implementations

We are aware of existing functionality in both `DropletUtils` and
`SpatialExperiment`. We are working with the authors of those packages to cover
the use cases in both those packages and possibly port I/O functionality into
`TENxIO`. We are using long tests and the `DropletTestFiles` package to
cover example datasets on `ExperimentHub`, if you would like to know more, see
the `longtests` directory on GitHub.

# Installation

```{r,eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("waldronlab/TENxIO")
```

# Load the package

```{r,include=TRUE,results="hide",message=FALSE,warning=FALSE}
library(TENxIO)
```

# Description

`TENxIO` offers an set of classes that allow users to easily work with files
typically obtained from the 10X Genomics website. Generally, these are outputs
from the Cell Ranger pipeline.

# Procedure

Loading the data into a Bioconductor class is a two step process. First, the
file must be identified by either the user or the `TENxFile` function. The
appropriate function will be evoked to provide a `TENxIO` class representation,
e.g., `TENxH5` for HDF5 files with an `.h5` extension. Secondly, the `import`
method for that particular file class will render a common Bioconductor class
representation for the user. The main representations used by the package are
`SingleCellExperiment`, `SummarizedExperiment`, `GRanges`, and
`RaggedExperiment`.

# Dataset versioning

The versioning schema in the package mostly applies to HDF5 resources and is
loosely based on versions of 10X datasets. For the most part, version 3 datasets
usually contain ranged information at specific locations in the data file.
Version 2 datasets will usually contain a `genes.tsv` file, rather than
`features.tsv` as in version 3. If the file version is unknown, the software
will attempt to derive the version from the data where possible.

# File classes

## TENxFile

The `TENxFile` class is the catch-all class superclass that allows transition
to subclasses pertinent to specific files. It inherits from the `BiocFile`
class and allows for easy dispatching `import` methods.

```{r}
showClass("TENxFile")
```

### `ExperimentHub` resources

`TENxFile` can handle resources from `ExperimentHub` with careful inputs.
For example, one can import a `TENxBrainData` dataset via the appropriate
`ExperimentHub` identifier (`EH1039`):

```{r}
hub <- ExperimentHub::ExperimentHub()
hub["EH1039"]
```

Currently, `ExperimentHub` resources do not have an extension and it is best to
provide that to the `TENxFile` constructor function.

```{r,eval=FALSE}
fname <- hub[["EH1039"]]
TENxFile(fname, extension = "h5", group = "mm10", version = "2")
```

Note. `EH1039` is a large ~ 4GB file and files without extension as those
obtained from `ExperimentHub` will emit a warning so that the user is aware that
the import operation may fail, esp. if the internal structure of the file is
modified.

### TENxH5

`TENxIO` mainly supports version 3 and 2 type of H5 files. These are files
with specific groups and names as seen in `h5.version.map`, an internal
`data.frame` map that guides the import operations.

```{r}
TENxIO:::h5.version.map
```

In the case that, there is a file without genomic coordinate information, the
constructor function can take an `NA_character_` input for the `ranges`
argument.

The `TENxH5` constructor function can be used on either version of these H5
files. In this example, we use a subset of the PBMC granulocyte
H5 file obtained from the [10X website](https://cf.10xgenomics.com/samples/cell-arc/2.0.0/pbmc_granulocyte_sorted_3k/pbmc_granulocyte_sorted_3k_filtered_feature_bc_matrix.h5).

```{r}
h5f <- system.file(
    "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)
library(rhdf5)
h5ls(h5f)
```

Note. The `h5ls` function gives an overview of the structure of the file.
It matches version 3 in our version map.

The show method gives an overview of the data components in the file:

```{r}
con <- TENxH5(h5f)
con
```

### import TENxH5 method

We can simply use the import method to convert the file representation to
a Bioconductor class representation, typically a `SingleCellExperiment`.

```{r}
import(con)
```

**Note**. Although the main representation in the package is
`SingleCellExperiment`, there could be a need for alternative data class
representations of the data. The `projection` field in the `TENxH5` show method
is an initial attempt to allow alternative representations.

## TENxMTX

Matrix Market formats are also supported (`.mtx` extension). These are typically
imported as SummarizedExperiment as they usually contain count data.

```{r}
mtxf <- system.file(
    "extdata", "pbmc_3k_ff_bc_ex.mtx",
    package = "TENxIO", mustWork = TRUE
)
con <- TENxMTX(mtxf)
con
```

## import MTX method

The `import` method yields a `SummarizedExperiment` without colnames or
rownames.

```{r}
import(con)
```

## TENxFileList

Generally, the 10X website will provide tarballs (with a `.tar.gz`
extension) which can be imported with the `TENxFileList` class. The tarball
can contain components of a gene expression experiment including the matrix
data, row data (aka 'features') expressed as Ensembl identifiers, gene symbols,
etc. and barcode information for the columns.

The `TENxFileList` class allows importing multiple files within a `tar.gz`
archive. The `untar` function with the `list = TRUE` argument shows all the file
names in the tarball.

```{r}
fl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
    package = "TENxIO", mustWork = TRUE
)
untar(fl, list = TRUE)
```

We then use the `import` method across all file types to obtain an integrated
Bioconductor representation that is ready for analysis. Files in `TENxFileList`
can be represented as a `SingleCellExperiment` with row names and column names.

```{r}
con <- TENxFileList(fl)
import(con)
```

## TENxPeaks

Peak files can be handled with the `TENxPeaks` class. These files are usually
named `*peak_annotation` files with a `.tsv` extension. Peak files are
represented as `GRanges`.

```{r}
pfl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv",
    package = "TENxIO", mustWork = TRUE
)
tenxp <- TENxPeaks(pfl)
peak_anno <- import(tenxp)
peak_anno
```

## TENxFragments

Fragment files are quite large and we make use of the `Rsamtools` package to
import them with the `yieldSize` parameter. By default, we use a `yieldSize` of
200.

```{r}
fr <- system.file(
    "extdata", "pbmc_3k_atac_ex_fragments.tsv.gz",
    package = "TENxIO", mustWork = TRUE
)
```

Internally, we use the `TabixFile` constructor function to work with indexed
`tsv.gz` files.

**Note**. A warning is emitted whenever a `yieldSize` parameter is not set.

```{r}
tfr <- TENxFragments(fr)
tfr
```

Because there may be a variable number of fragments per barcode, we use a
`RaggedExperiment` representation for this file type.

```{r}
fra <- import(tfr)
fra
```

Similar operations to those used with `SummarizedExperiment` are supported. For
example, the genomic ranges can be displayed via `rowRanges`:

```{r}
rowRanges(fra)
```

# Session Information

```{r session-info, echo=FALSE, results='asis'}
html_output <- paste0(
    "<details><summary>",
    "Click to expand <code>sessionInfo()</code>",
    "</summary>\n\n",
    "<pre><code>",
    paste(capture.output(sessionInfo()), collapse = "\n"),
    "</code></pre>\n</details>"
)
cat(html_output)
```

---
title: "TENxIO: Import Single Cell Data Files"
author: "Marcel Ramos"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{TENxIO Quick Start Guide}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    number_sections: no
    toc: yes
    toc_depth: 4
package: TENxIO
---

<!-- badges: start -->
<!-- badges: end -->

```{r, setup, include=FALSE}
knitr::opts_chunk$set(cache = TRUE)
```

# Introduction

`TENxIO` allows users to import 10X pipeline files into known Bioconductor
classes. The package is not comprehensive, there are file types that are not
supported. For Visium datasets, we direct users to the `VisiumIO` package on
Bioconductor. TENxIO consolidates functionality from `DropletUtils`. If you
would like a file format to be supported, open an issue at
https://github.com/waldronlab/TENxIO.

# Supported Formats

| **Extension**       | **Class**     | **Imported as**      |
|---------------------|---------------|----------------------|
| .h5                 | TENxH5        | SingleCellExperiment w/ TENxMatrix |
| .mtx / .mtx.gz      | TENxMTX       | SummarizedExperiment w/ dgCMatrix |
| .tar.gz             | TENxFileList  | SingleCellExperiment w/ dgCMatrix |
| peak_annotation.tsv | TENxPeaks     | GRanges              |
| fragments.tsv.gz    | TENxFragments | RaggedExperiment     |
| .tsv / .tsv.gz      | TENxTSV       | tibble               |
| spatial.tar.gz      | TENxSpatialList | inter. DataFrame list |

# Tested 10X Products

We have tested these functions with _some_
[datasets](https://www.10xgenomics.com/resources/datasets) from 10x Genomics
including those from:

* Single Cell Gene Expression
* Single Cell ATAC
* Single Cell Multiome ATAC + Gene Expression
* Spatial Gene Expression

Note. That extensive testing has not been performed and the codebase may require
some adaptation to ensure compatibility with all pipeline outputs.

# Bioconductor implementations

We are aware of existing functionality in both `DropletUtils` and
`SpatialExperiment`. We are working with the authors of those packages to cover
the use cases in both those packages and possibly port I/O functionality into
`TENxIO`. We are using long tests and the `DropletTestFiles` package to
cover example datasets on `ExperimentHub`, if you would like to know more, see
the `longtests` directory on GitHub.

# Installation

```{r, install, eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("waldronlab/TENxIO")
```

# Load the package

```{r, library, include=TRUE,results="hide",message=FALSE,warning=FALSE}
library(TENxIO)
```

# Description

`TENxIO` offers an set of classes that allow users to easily work with files
typically obtained from the 10X Genomics website. Generally, these are outputs
from the Cell Ranger pipeline.

# Procedure

Loading the data into a Bioconductor class is a two step process. First, the
file must be identified by either the user or the `TENxFile` function. The
appropriate function will be evoked to provide a `TENxIO` class representation,
e.g., `TENxH5` for HDF5 files with an `.h5` extension. Secondly, the `import`
method for that particular file class will render a common Bioconductor class
representation for the user. The main representations used by the package are
`SingleCellExperiment`, `SummarizedExperiment`, `GRanges`, and
`RaggedExperiment`.

# Dataset versioning

The versioning schema in the package mostly applies to HDF5 resources and is
loosely based on versions of 10X datasets. For the most part, version 3 datasets
usually contain ranged information at specific locations in the data file.
Version 2 datasets will usually contain a `genes.tsv` file, rather than
`features.tsv` as in version 3. If the file version is unknown, the software
will attempt to derive the version from the data where possible.

# File classes

## TENxFile

The `TENxFile` class is the catch-all class superclass that allows transition
to subclasses pertinent to specific files. It inherits from the `BiocFile`
class and allows for easy dispatching `import` methods.

```{r, show-TENxFile-class}
showClass("TENxFile")
```

### `ExperimentHub` resources

`TENxFile` can handle resources from `ExperimentHub` with careful inputs.
For example, one can import a `TENxBrainData` dataset via the appropriate
`ExperimentHub` identifier (`EH1039`):

```{r, hub-resource}
hub <- ExperimentHub::ExperimentHub()
hub["EH1039"]
```

Currently, `ExperimentHub` resources do not have an extension and it is best to
provide that to the `TENxFile` constructor function.

```{r, TENxFile-hub, eval=FALSE}
fname <- hub[["EH1039"]]
TENxFile(fname, extension = "h5", group = "mm10", version = "2")
```

Note. `EH1039` is a large ~ 4GB file and files without extension as those
obtained from `ExperimentHub` will emit a warning so that the user is aware that
the import operation may fail, esp. if the internal structure of the file is
modified.

### TENxH5

`TENxIO` mainly supports version 3 and 2 type of H5 files. These are files
with specific groups and names as seen in `h5.version.map`, an internal
`data.frame` map that guides the import operations.

```{r, h5-version-map}
TENxIO:::h5.version.map
```

In the case that, there is a file without genomic coordinate information, the
constructor function can take an `NA_character_` input for the `ranges`
argument.

The `TENxH5` constructor function can be used on either version of these H5
files. In this example, we use a subset of the PBMC granulocyte
H5 file obtained from the [10X website](https://cf.10xgenomics.com/samples/cell-arc/2.0.0/pbmc_granulocyte_sorted_3k/pbmc_granulocyte_sorted_3k_filtered_feature_bc_matrix.h5).

```{r, h5ls}
h5f <- system.file(
    "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)
library(rhdf5)
h5ls(h5f)
```

Note. The `h5ls` function gives an overview of the structure of the file.
It matches version 3 in our version map.

The show method gives an overview of the data components in the file:

```{r, TENxH5-constructor}
con <- TENxH5(h5f)
con
```

### import TENxH5 method

We can simply use the import method to convert the file representation to
a Bioconductor class representation, typically a `SingleCellExperiment`.

```{r, import-TENxH5}
import(con)
```

**Note**. Although the main representation in the package is
`SingleCellExperiment`, there could be a need for alternative data class
representations of the data. The `projection` field in the `TENxH5` show method
is an initial attempt to allow alternative representations.

## TENxMTX

Matrix Market formats are also supported (`.mtx` extension). These are typically
imported as SummarizedExperiment as they usually contain count data.

```{r, TENxMTX-constructor}
mtxf <- system.file(
    "extdata", "pbmc_3k_ff_bc_ex.mtx",
    package = "TENxIO", mustWork = TRUE
)
con <- TENxMTX(mtxf)
con
```

## import MTX method

The `import` method yields a `SummarizedExperiment` without colnames or
rownames.

```{r, import-TENxMTX}
import(con)
```

## TENxFileList

Generally, the 10X website will provide tarballs (with a `.tar.gz`
extension) which can be imported with the `TENxFileList` class. The tarball
can contain components of a gene expression experiment including the matrix
data, row data (aka 'features') expressed as Ensembl identifiers, gene symbols,
etc. and barcode information for the columns.

The `TENxFileList` class allows importing multiple files within a `tar.gz`
archive. The `untar` function with the `list = TRUE` argument shows all the file
names in the tarball.

```{r, TENxFileList-list}
fl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
    package = "TENxIO", mustWork = TRUE
)
untar(fl, list = TRUE)
```

We then use the `import` method across all file types to obtain an integrated
Bioconductor representation that is ready for analysis. Files in `TENxFileList`
can be represented as a `SingleCellExperiment` with row names and column names.

```{r, import-TENxFileList}
con <- TENxFileList(fl)
import(con)
```

## TENxPeaks

Peak files can be handled with the `TENxPeaks` class. These files are usually
named `*peak_annotation` files with a `.tsv` extension. Peak files are
represented as `GRanges`.

```{r, TENxPeaks-import}
pfl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv",
    package = "TENxIO", mustWork = TRUE
)
tenxp <- TENxPeaks(pfl)
peak_anno <- import(tenxp)
peak_anno
```

## TENxFragments

Fragment files are quite large and we make use of the `Rsamtools` package to
import them with the `yieldSize` parameter. By default, we use a `yieldSize` of
200.

```{r, TENxFragments-path}
fr <- system.file(
    "extdata", "pbmc_3k_atac_ex_fragments.tsv.gz",
    package = "TENxIO", mustWork = TRUE
)
```

Internally, we use the `TabixFile` constructor function to work with indexed
`tsv.gz` files.

**Note**. A warning is emitted whenever a `yieldSize` parameter is not set.

```{r, TENxFragments-constructor}
tfr <- TENxFragments(fr)
tfr
```

Because there may be a variable number of fragments per barcode, we use a
`RaggedExperiment` representation for this file type.

```{r, import-TENxFragments}
fra <- import(tfr)
fra
```

Similar operations to those used with `SummarizedExperiment` are supported. For
example, the genomic ranges can be displayed via `rowRanges`:

```{r, rowRanges-TENxFragments}
rowRanges(fra)
```

# Session Information

```{r, sessioninfo, echo=FALSE, results='asis'}
html_output <- paste0(
    "<details><summary>",
    "Click to expand <code>sessionInfo()</code>",
    "</summary>\n\n",
    "<pre><code>",
    paste(capture.output(sessionInfo()), collapse = "\n"),
    "</code></pre>\n</details>"
)
cat(html_output)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFileList-class.R
\name{TENxFileList}
\alias{TENxFileList}
\title{TENxFileList: Represent groups of files from 10X Genomic}
\usage{
TENxFileList(..., version, compressed = FALSE)
}
\arguments{
\item{...}{Typically, a file path to a tarball archive. Can be named
arguments corresponding to file paths, or a named list of file paths.}

\item{version}{character(1) The version in the tarball. See details.}

\item{compressed}{logical(1) Whether or not the file provided is compressed,
usually as \code{tar.gz} (default FALSE)}
}
\value{
Either a \code{SingleCellExperiment} or a list of imported data
}
\description{
This constructor function is meant to handle \code{.tar.gz} tarball
files from 10X Genomics.
}
\details{
These tarballs usually contain three files:
\enumerate{
\item \code{matrix.mtx.gz} - the counts matrix
\item \code{features.tsv.gz} - row metadata usually represented as \code{rowData}
\item \code{barcodes.tsv.gz} - column names corresponding to cell barcode
identifiers
If all above files are in the tarball, the import method will provide a
\code{SingleCellExperiment}. Otherwise, a simple list of imported data is given.
Note that version "3" uses 'features.tsv.gz' and version "2" uses
'genes.tsv.gz'. If known, indicate the \code{version} argument in the
\code{TENxFileList} constructor function.
}
}
\examples{

fl <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ff_bc_ex_matrix.tar.gz",
    package = "TENxIO", mustWork = TRUE
)

## Method 1 (tarball)
TENxFileList(fl)

## metadata before import
metadata(TENxFileList(fl))

## import() method
import(TENxFileList(fl))

## metadata after import
import(TENxFileList(fl)) |>
    metadata()

## untar to simulate folder output
dir.create(tdir <- tempfile())
untar(fl, exdir = tdir)

## Method 2 (folder)
TENxFileList(tdir)
import(TENxFileList(tdir))

## Method 3 (list of TENxFile objects)
files <- list.files(tdir, recursive = TRUE, full.names = TRUE)
names(files) <- basename(files)
filelist <- lapply(files, TENxFile)

TENxFileList(filelist, compressed = FALSE)

## Method 4 (SimpleList)
TENxFileList(as(filelist, "SimpleList"), compressed = FALSE)

## Method 5 (named arguments)
TENxFileList(
    barcodes.tsv.gz = TENxFile(files[1]),
    features.tsv.gz = TENxFile(files[2]),
    matrix.mtx.gz = TENxFile(files[3])
)

unlink(tdir, recursive = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxMTX-class.R
\name{TENxMTX}
\alias{TENxMTX}
\title{TENxMTX: Represent Matrix Market Format Files from 10X}
\usage{
TENxMTX(resource, compressed = FALSE, ...)
}
\arguments{
\item{resource}{character(1) The path to the file}

\item{compressed}{logical(1) Whether the resource file is compressed (default
FALSE)}

\item{...}{Additional inputs to the low level class generator functions}
}
\value{
A \code{SummarizedExperiment} instance with a \code{dgCMatrix} in the assay
}
\description{
This constructor function accepts \code{.mtx} and \code{.mtx.gz} compressed formats
for eventual importing. It is mainly used with tarball files from 10X
Genomics, where more annotation data is included. Importing solely the
\code{.mtx} format will provide users with a \code{SummarizedExperiment} with an assay
of class \code{dgCMatrix} from the \code{Matrix} package. Currently, other formats are
not supported but if you'd like to request support for a format, please open
an issue on GitHub.
}
\examples{

mtxf <- system.file(
    "extdata", "pbmc_3k_ff_bc_ex.mtx",
    package = "TENxIO", mustWork = TRUE
)

con <- TENxMTX(mtxf)

import(con)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFile-class.R
\name{TENxFile}
\alias{TENxFile}
\title{TENxFile constructor function}
\usage{
TENxFile(resource, extension, ...)
}
\arguments{
\item{resource}{character(1) The path to the file}

\item{extension}{character(1) The file extension for the given resource. It
can usually be obtained from the file path. An override can be provided
especially for \code{ExperimentHub} resources where the file extension is
removed.}

\item{...}{Additional inputs to the low level class generator functions}
}
\value{
A subclass of \code{TENxFile} according to the input file extension
}
\description{
The \code{TENxFile} constructor function serves as the
auto-recognizer function for 10X files. It can import several different
file extensions, namely:
\preformatted{
    * H5 - on-disk HDF5
    * MTX - matrix market
    * .tar.gz - compressed tarball
}
}
\details{
\strong{Note} that the example below includes the use of a large ~ 4 GB
\code{ExperimentHub} resource obtained from the 10X website.
}
\examples{
if (interactive()) {

    ## from ExperimentHub
    hub <- ExperimentHub::ExperimentHub()
    fname <- hub[["EH1039"]]
    TENxFile(fname, extension = "h5", group = "mm10", version = "2")
    TENxFile(fname, extension = "h5", group = "mm10", version = "2") |>
        metadata()

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxPeaks-class.R
\name{TENxPeaks}
\alias{TENxPeaks}
\title{Import 10x peak annotation files from 10x}
\usage{
TENxPeaks(resource, extension, ...)
}
\arguments{
\item{resource}{character(1) The path to the file}

\item{extension}{character(1) The file extension for the given resource. It
can usually be obtained from the file path. An override can be provided
especially for \code{ExperimentHub} resources where the file extension is
removed.}

\item{...}{Additional inputs to the low level class generator functions}
}
\value{
A \code{GRanges} class object of peak locations
}
\description{
This constructor function is designed to work with the files denoted with
"peak_annotation" in the file name. These are usually produced as tab
separated value files, i.e., \code{.tsv}.
}
\details{
The output class allows handling of peak data. It can be used in
conjunction with the \code{annotation} method on a \code{SingleCellExperiment} to add
peak information to the experiment. The ranged data is represented as a
\code{GRanges} class object.
}
\examples{

fi <- system.file(
    "extdata", "pbmc_granulocyte_sorted_3k_ex_atac_peak_annotation.tsv",
    package = "TENxIO", mustWork = TRUE
)
peak_file <- TENxPeaks(fi)
peak_anno <- import(peak_file)
peak_anno

example(TENxH5)

## Add peaks to an existing SCE
## First, import the SCE from an example H5 file
h5f <- system.file(
    "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)
con <- TENxH5(h5f)
sce <- import(con)
## auto-import peaks when using annotation<-
annotation(sce, name = "peak_annotation") <- peak_file
annotation(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxMTX-class.R
\docType{class}
\name{TENxMTX-class}
\alias{TENxMTX-class}
\alias{.TENxMTX}
\alias{import,TENxMTX,ANY,ANY-method}
\title{TENxMTX: The Matrix Market representation class for 10X Data}
\usage{
\S4method{import}{TENxMTX,ANY,ANY}(con, format, text, ...)
}
\arguments{
\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{...}{Parameters to pass to the format-specific method.}
}
\value{
A \code{TENxMTX} class object
}
\description{
This class is designed to work with 10x MTX datasets,
particularly from the multiome pipelines.
}
\details{
The \code{TENxMTX} class is a straightforward implementation that allows
the user to import a Matrix Market file format using \code{Matrix::readMM}.
Currently, it returns a \code{SummarizedExperiment} with an internal \code{dgCMatrix}
assay. To request other formats, please open an issue on GitHub.
}
\section{Methods (by generic)}{
\itemize{
\item \code{import(con = TENxMTX, format = ANY, text = ANY)}: Import method mainly for mtx.gz files from 10x

}}
\section{Slots}{

\describe{
\item{\code{compressed}}{logical(1) Whether or not the file is in compressed format,
usually gzipped (\code{.gz}).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFragments-class.R
\docType{class}
\name{TENxFragments-class}
\alias{TENxFragments-class}
\alias{.TENxFragments}
\alias{import,TENxFragments,ANY,ANY-method}
\title{TENxFragments: A class to represent fragments data as \code{GRanges}}
\usage{
\S4method{import}{TENxFragments,ANY,ANY}(con, format, text, ...)
}
\arguments{
\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{...}{Parameters to pass to the format-specific method.}
}
\value{
A \code{TENxFragments} class object
}
\description{
This class is designed to work mainly with \code{fragments.tsv.gz} files from
10x pipelines.
}
\details{
Fragments data from 10x can be quite large. In order to speed up
the initial exploration of the data, we use a default of \strong{200} records
for loading. Users can change this default value by specifying a new one
via the \code{yieldSize} argument in the constructor function.
}
\section{Methods (by generic)}{
\itemize{
\item \code{import(con = TENxFragments, format = ANY, text = ANY)}: Import method for representing fragments.tsv.gz
data from 10x via \code{Rsamtools} and \code{RaggedExperiment}

}}
\section{Slots}{

\describe{
\item{\code{which}}{GRanges() A GRanges indicating the regions of interest. This
get sent to \code{RSamtools} as the \code{param} input.}

\item{\code{yieldSize}}{numeric() The number of records to read by default, 200
records will be imported. A warning will be emitted if not modified.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxTSV-class.R
\docType{class}
\name{TENxTSV-class}
\alias{TENxTSV-class}
\alias{.TENxTSV}
\alias{TENxTSV}
\alias{import,TENxTSV,ANY,ANY-method}
\alias{metadata,TENxTSV-method}
\title{TENxTSV: A class to represent 10x tab separated values files}
\usage{
\S4method{import}{TENxTSV,ANY,ANY}(con, format, text, ...)

TENxTSV(resource, compressed = FALSE, ...)

\S4method{metadata}{TENxTSV}(x, ...)
}
\arguments{
\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{...}{Parameters to pass to the format-specific method.}

\item{resource}{character(1) The path to the file}

\item{compressed}{logical(1) Whether the resource file is compressed (default
FALSE)}

\item{x}{A \code{TENxTSV} object}
}
\value{
A \code{TENxTSV} class object; a \code{tibble} for the import method
}
\description{
This class is general purpose for reading in tabular data from
the 10x Genomics website with the \code{.tsv} file extension. The class also
supports compressed files, i.e., those with the \code{.tsv.gz} extension.
}
\details{
Typical \code{.tsv} files obtained from the 10X website are compressed
and contain information relevant to 'barcodes' and 'features'. Currently,
the code only supports files such as \verb{features.tsv.*} and \verb{barcodes.tsv.*}.
}
\section{Functions}{
\itemize{
\item \code{import(con = TENxTSV, format = ANY, text = ANY)}: General import method for \code{tsv} files from 10x;
using \code{readr::read_tsv} and returning a \code{tibble} representation

\item \code{metadata(TENxTSV)}: \code{metadata} method for \code{TENxTSV} objects

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxH5-class.R
\docType{class}
\name{TENxH5-class}
\alias{TENxH5-class}
\alias{.TENxH5}
\alias{rowData,TENxH5-method}
\alias{dim,TENxH5-method}
\alias{dimnames,TENxH5-method}
\alias{genome,TENxH5-method}
\alias{rowRanges,TENxH5-method}
\alias{import,TENxH5,ANY,ANY-method}
\alias{show,TENxH5-method}
\title{TENxH5: The HDF5 file representation class for 10X Data}
\usage{
\S4method{rowData}{TENxH5}(x, use.names = TRUE, ...)

\S4method{dim}{TENxH5}(x)

\S4method{dimnames}{TENxH5}(x)

\S4method{genome}{TENxH5}(x)

\S4method{rowRanges}{TENxH5}(x, ...)

\S4method{import}{TENxH5,ANY,ANY}(con, format, text, ...)

\S4method{show}{TENxH5}(object)
}
\arguments{
\item{x}{A \code{TENxH5} object}

\item{use.names}{
    For \code{rowData}: Like \code{\link[S4Vectors]{mcols}(x)}, by default
    \code{rowData(x)} propagates the rownames of \code{x} to the returned
    \link[S4Vectors]{DataFrame} object (note that for a SummarizedExperiment
    object or derivative, the rownames are also the names i.e.
    \code{rownames(x)} is always the same as \code{names(x)}).
    Setting \code{use.names=FALSE} suppresses this propagation i.e. it
    returns a \link[S4Vectors]{DataFrame} object with no rownames. Use this
    when \code{rowData(x)} fails, which can happen when the rownames contain
    NAs (because the rownames of a SummarizedExperiment object or derivative
    can contain NAs, but the rownames of a \link[S4Vectors]{DataFrame} object
    cannot).

    For \code{combineRows} and \code{combineCols}: See Combining section
    below.
  }

\item{...}{
    For \code{assay}, arguments in \code{...} are forwarded to \code{assays}.

    For \code{rbind}, \code{cbind}, \code{...} contains SummarizedExperiment
    objects (or derivatives) to be combined.

    For other accessors, ignored.
  }

\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{object}{A \code{TENxH5} class object}
}
\value{
A \code{TENxH5} class object
}
\description{
This class is designed to work with 10x Single Cell datasets.
It was developed using the PBMC 3k 10X dataset from the CellRanger v2
pipeline.
}
\details{
The data version "3" mainly includes a "matrix" group and "interval"
information within the file. Version "2" data does not include
ranged-based information and has a different directory structure compared
to version "3". See the internal \code{data.frame}: \code{TENxIO:::h5.version.map} for
a map of fields and their corresponding file locations within the H5 file.
This map is used to create the \code{rowData} structure from the file.
}
\section{Methods (by generic)}{
\itemize{
\item \code{rowData(TENxH5)}: Generate the rowData ad hoc from a TENxH5 file

\item \code{dim(TENxH5)}: Get the dimensions of the data as stored in the file

\item \code{dimnames(TENxH5)}: Get the dimension names from the file

\item \code{genome(TENxH5)}: Read genome string from file

\item \code{rowRanges(TENxH5)}: Read interval data and represent as GRanges

\item \code{import(con = TENxH5, format = ANY, text = ANY)}: Import TENxH5 data as a SingleCellExperiment; see section
below

\item \code{show(TENxH5)}: Display a snapshot of the contents within a TENxH5 file
before import

}}
\section{Slots}{

\describe{
\item{\code{version}}{character(1) There are currently two recognized versions
associated with 10X data, either version "2" or "3". See details for more
information.}

\item{\code{group}}{character(1) The HDF5 group embedded within the file structure,
this is usually either the "matrix" or "outs" group but other groups are
supported as well.}

\item{\code{ranges}}{character(1) The HDF5 internal folder location embedded within
the file that points to the ranged data information, e.g.,
"/features/interval".}
}}

\section{import}{

The \code{import} method uses \code{DelayedArray::TENxMatrix} to represent matrix
data. Generally, version 3 datasets contain associated genomic coordinates.
The associated feature data, as displayed by the \code{rowData} method, is
queried for the "Type" column which will indicate that a \code{splitAltExps}
operation is appropriate. If a \code{ref} input is provided to the constructor
function \code{TENxH5}, it will be used as the main experiment; otherwise, the
most frequent category in the "Type" column will be used. For example,
the Multiome ATAC + Gene Expression feature data contains both 'Gene
Expression' and 'Peaks' labels in the "Type" column.
}

\seealso{
\link{TENxH5}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxIO-package.R
\docType{package}
\name{TENxIO}
\alias{TENxIO}
\alias{TENxIO-package}
\title{TENxIO: A Bioconductor package for importing 10X Genomics files}
\description{
The package provides file classes based on \code{BiocIO} for common file
extensions found in the 10X Genomics website.
}
\section{Supported file types}{

Here is a table of supported file and file extensions and their imported
classes:\tabular{lll}{
   \strong{Extension} \tab \strong{Class} \tab \strong{Imported as} \cr
   .h5 \tab TENxH5 \tab SingleCellExperiment w/ TENxMatrix \cr
   .mtx / .mtx.gz \tab TENxMTX \tab SummarizedExperiment w/ dgCMatrix \cr
   .tar.gz \tab TENxFileList \tab SingleCellExperiment w/ dgCMatrix \cr
   peak_annotation.tsv \tab TENxPeaks \tab GRanges \cr
   fragments.tsv.gz \tab TENxFragments \tab RaggedExperiment \cr
   .tsv / .tsv.gz \tab TENxTSV \tab tibble \cr
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/waldronlab/TENxIO}
  \item Report bugs at \url{https://github.com/waldronlab/TENxIO/issues}
}

}
\author{
\strong{Maintainer}: Marcel Ramos \email{marcel.ramos@sph.cuny.edu} (\href{https://orcid.org/0000-0002-3242-0582}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFragments-class.R
\name{TENxFragments}
\alias{TENxFragments}
\title{TENxFragments: Import fragments files from 10X}
\usage{
TENxFragments(resource, yieldSize = 200, which = GRanges(), ...)
}
\arguments{
\item{resource}{character(1) The file path to the fragments resource, usually
a compressed tabix file with extension \code{.tsv.gz}.}

\item{yieldSize}{numeric() The number of records to read by default, 200
records will be imported. A warning will be emitted if not modified.}

\item{which}{GRanges() A GRanges indicating the regions of interest. This
get sent to \code{RSamtools} as the \code{param} input.}

\item{...}{Further arguments to the class generator function (currently not
used)}
}
\value{
A \code{RaggedExperiment} object class
}
\description{
TENxFragments: Import fragments files from 10X
}
\examples{

fr <- system.file(
    "extdata", "pbmc_3k_atac_ex_fragments.tsv.gz",
    package = "TENxIO", mustWork = TRUE
)

tfr <- TENxFragments(fr)

fra <- import(tfr)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxPeaks-class.R
\docType{class}
\name{TENxPeaks-class}
\alias{TENxPeaks-class}
\alias{.TENxPeaks}
\alias{import,TENxPeaks,ANY,ANY-method}
\alias{annotation<-,SingleCellExperiment,ANY-method}
\alias{annotation,SingleCellExperiment-method}
\title{TENxPeaks: The class to represent 10x Peaks files}
\usage{
\S4method{import}{TENxPeaks,ANY,ANY}(con, format, text, ...)

\S4method{annotation}{SingleCellExperiment,ANY}(object, ...) <- value

\S4method{annotation}{SingleCellExperiment}(object, ...)
}
\arguments{
\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{...}{Parameters to pass to the format-specific method.}

\item{object}{The object to export.}

\item{value}{
    The annotation information to set on \code{object}.
  }
}
\value{
A \code{TENxPeaks} class object
}
\description{
This class is designed to work with the files denoted with "peak_annotation"
in the file name. These are usually produced as tab separated value files,
i.e., \code{.tsv}.
}
\details{
This class is a straightforward class for handling peak data. It can
be used in conjunction with the \code{annotation} method on a
\code{SingleCellExperiment} to add peak information to the experiment. The
ranged data is represented as a \code{GRanges} class object.
}
\section{Functions}{
\itemize{
\item \code{import(con = TENxPeaks, format = ANY, text = ANY)}: Import a peaks_annotation file from 10x as a
\code{GRanges} representation

\item \code{annotation(object = SingleCellExperiment) <- value}: Replacement method to add annotation data to a
\code{SingleCellExperiment}

\item \code{annotation(SingleCellExperiment)}: Extraction method to obtain annotation data from
a \code{SingleCellExperiment} representation

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxH5-class.R
\name{TENxH5}
\alias{TENxH5}
\title{TENxH5: Represent H5 files from 10X}
\usage{
TENxH5(resource, version, group, ranges, rowidx, colidx, ...)
}
\arguments{
\item{resource}{character(1) The path to the file}

\item{version}{character(1) There are currently two recognized versions
associated with 10X data, either version "2" or "3". See details for more
information.}

\item{group}{character(1) The HDF5 group embedded within the file structure,
this is usually either the "matrix" or "outs" group but other groups are
supported as well (e.g., "mm10").}

\item{ranges}{character(1) The HDF5 internal folder location embedded within
the file that points to the ranged data information, e.g.,
"/features/interval". Set to \code{NA_character_} if range information is not
present.}

\item{rowidx, colidx}{numeric() A vector of indices corresponding to either
rows or columns that will dictate the data imported from the file. The
indices will be passed on to the \code{[} method of the \code{TENxMatrix}
representation.}

\item{...}{Additional inputs to the low level class generator functions}
}
\value{
Usually, a \code{SingleCellExperiment} instance
}
\description{
This constructor function was developed using the PBMC 3K dataset from 10X
Genomics (version 3). Other versions are supported and input arguments
\code{version} and \code{group} can be overridden.
}
\details{
The various \code{TENxH5} methods including \code{rowData} and \code{rowRanges},
provide a snapshot of the data using a length 12 head and tail subset for
efficiency. In contrast, methods such as \code{dimnames} and \code{dim} give
a full view of the dimensions of the data. The \code{show} method provides
relevant information regarding the dimensions of the data including
metadata such as \code{rowData} and "Type" column, if available. The data files
are \code{import}ed as \code{SingleCellExperiment} class objects.

An additional \code{ref} argument can be provided when the file contains
multiple \code{feature_type} in the file or "Type" in the \code{rowData}. By default,
the first type reported in \code{table()} is set as the \code{mainExpName} in the
\code{SingleCellExperiment} object.

For data that do not contain genomic coordinate information, the \code{TENxH5}
will fail to read \code{"/features/interval"} and will set the \code{ranges}
argument to \code{NA_character_}.

The data version "3" mainly includes a "matrix" group and "interval"
information within the file. Version "2" data does not include
ranged-based information and has a different directory structure compared
to version "3". See the internal \code{data.frame}: \code{TENxIO:::h5.version.map}
for a map of fields and their corresponding file locations within the H5
file. This map is used to create the \code{rowData} structure from the file.
}
\examples{

h5f <- system.file(
    "extdata", "pbmc_granulocyte_ff_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)

TENxH5(h5f)

import(TENxH5(h5f))

h5f <- system.file(
    "extdata", "10k_pbmc_ATACv2_f_bc_ex.h5",
    package = "TENxIO", mustWork = TRUE
)

## Optional ref input, most frequent Type used by default
th5 <- TENxH5(h5f, ranges = "/features/id", ref = "Peaks")
th5
TENxH5(h5f, ranges = "/features/id")
import(th5)

}
\seealso{
\code{import} section in \linkS4class{TENxH5}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFile-class.R
\docType{class}
\name{TENxFile-class}
\alias{TENxFile-class}
\alias{.TENxFile}
\alias{metadata,TENxFile-method}
\title{TENxFile: General purpose class for 10X files}
\usage{
\S4method{metadata}{TENxFile}(x, ...)
}
\arguments{
\item{x}{An object of class \code{TENxFile}, \code{TENxFileList}, \code{TENxMTX}, \code{TENxH5},
\code{TENxPeaks}, \code{TENxTSV}, or derivatives}

\item{...}{Additional arguments (not used)}
}
\value{
A list of metadata for the given object
}
\description{
The \code{TENxFile} class is the default representation for
unrecognized subclasses. It inherits from the BiocFile class and adds a few
additional slots. The constructor function can handle typical 10X file
types. For more details, see the constructor function documentation.
}
\section{Functions}{
\itemize{
\item \code{metadata(TENxFile)}: \code{metadata} method for \code{TENxFile} objects

}}
\section{Slots}{

\describe{
\item{\code{extension}}{character(1) The file extension as extracted from the file
path or overridden via the \code{ext} argument in the constructor function.}

\item{\code{colidx}}{integer(1) The column index corresponding to the columns in the
file that will subsequently be imported}

\item{\code{rowidx}}{integer(1) The row index corresponding to rows in the file that
will subsequently be imported}

\item{\code{remote}}{logical(1) Whether the file exists on the web, i.e., the
\code{resource} is a URL}

\item{\code{compressed}}{logical(1) Whether the file is compressed with, e.g., \code{.gz}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TENxFileList-class.R
\docType{class}
\name{TENxFileList-class}
\alias{TENxFileList-class}
\alias{.TENxFileList}
\alias{path,TENxFileList-method}
\alias{decompress,TENxFileList-method}
\alias{import,TENxFileList,ANY,ANY-method}
\alias{metadata,TENxFileList-method}
\title{TENxFileList: A list-like representation for TENxFiles}
\usage{
\S4method{path}{TENxFileList}(object, ...)

\S4method{decompress}{TENxFileList}(manager, con, ...)

\S4method{import}{TENxFileList,ANY,ANY}(con, format, text, ...)

\S4method{metadata}{TENxFileList}(x, ...)
}
\arguments{
\item{object}{
    An object containing paths. Even though it will typically contain
    a single path, \code{object} can actually contain an arbitrary
    number of paths.
  }

\item{...}{Additional arguments (not used)}

\item{manager}{A \code{ConnectionManager} internal instance; currently not used.}

\item{con}{The connection from which data is loaded or to which data is
saved. If this is a \code{character} vector, it is assumed to be a file name and
a corresponding file connection is created and then closed after exporting
the object. If it is a \link[BiocIO:BiocFile-class]{BiocFile} derivative, the
data is loaded from or saved to the underlying resource. If missing, the
function will return the output as a character vector, rather than writing
to a connection.}

\item{format}{The format of the output. If missing and \code{con} is a
file name, the format is derived from the file extension. This argument is
unnecessary when \code{con} is a derivative of \link[BiocIO:BiocFile-class]{BiocFile}.}

\item{text}{If \code{con} is missing, this can be a character vector directly
providing the string data to import.}

\item{x}{An object of class \code{TENxFile}, \code{TENxFileList}, \code{TENxMTX}, \code{TENxH5},
\code{TENxPeaks}, \code{TENxTSV}, or derivatives}
}
\value{
A \code{TENxFileList} class object
}
\description{
This class was designed to mainly handle tarballs from
10X Genomics. The typical file extension for these tarballs is \code{.tar.gz}.
}
\details{
These tarballs usually contain three files:
\enumerate{
\item \code{matrix.mtx.gz} - the counts matrix
\item \code{features.tsv.gz} - row metadata usually represented as \code{rowData}
\item \code{barcodes.tsv.gz} - column names corresponding to cell barcode
identifiers
Note that version '2' includes \code{genes.tsv.gz} instead of \code{features.tsv.gz} in
version '3'.
}

An additional \code{ref} argument can be provided when the file contains
multiple \code{feature_type} in the file or "Type" in the \code{rowData}. By default,
the first type reported in \code{table()} is set as the \code{mainExpName} in the
\code{SingleCellExperiment} object.
}
\section{Functions}{
\itemize{
\item \code{path(TENxFileList)}: Obtain file paths for all files in the object
as a vector

\item \code{decompress(TENxFileList)}: An intermediate method for decompressing
(via untar) the contents of a \code{.tar.gz} file list

\item \code{import(con = TENxFileList, format = ANY, text = ANY)}: Recursively import files within a
\code{TENxFileList}

\item \code{metadata(TENxFileList)}: \code{metadata} method for \code{TENxFileList} objects

}}
\section{Slots}{

\describe{
\item{\code{listData}}{list() The data in list format}

\item{\code{extension}}{character() A vector of file extensions for each file}

\item{\code{compressed}}{logical(1) Whether the file is compressed as \code{.tar.gz}}

\item{\code{version}}{character(1) The version number of the tarball usually either
'2' or '3'}
}}
